from ._mapi import MidasAPI




class Material:
    mats = []
    ids = []
    def __init__(self,data,id=None):
        if id == None: id =0
        if Material.ids == []: 
            count = 1
        else:
            count = max(Material.ids)+1
        if id == 0 or id in Material.ids: self.ID = count
        if id!= 0 and id not in Material.ids: self.ID = id

        self.DATA = data

        Material.mats.append(self)
        Material.ids.append(self.ID)
    
    @classmethod
    def json(cls):
        json = {"Assign":{}}
        for k in cls.mats:
            json["Assign"][k.ID]=k.DATA
        return json
    
    @staticmethod
    def create_only():
        return MidasAPI("PUT","/db/MATL",Material.json())
        
    @staticmethod
    def get():
        return MidasAPI("GET","/db/MATL")
    
    
    @staticmethod
    def delete():
        MidasAPI("DELETE","/db/MATL")
        Material.clear()

    @staticmethod
    def clear():
        Material.mats=[]
        Material.ids=[]

    @staticmethod
    def sync():
        a = Material.get()
        if a != {'message': ''}:
            if list(a['MATL'].keys()) != []:
                Material.mats = []
                Material.ids=[]
                for j in a['MATL'].keys():
                    Material(a['MATL'][j], int(j))

        # ----------------------------------  ALL FUNCTIONS  ---------------------------------------------------
    
    @staticmethod
    def create():
        if Material.mats!=[] : Material.create_only()
        if CreepShrinkage.mats!=[] : CreepShrinkage.create()
        if CompStrength.mats!=[] : CompStrength.create()
        if TDMatLink.json()!={'Assign':{}} : TDMatLink.create()
        
    
    @staticmethod
    def deleteAll():
        Material.delete()
        CreepShrinkage.delete()
        CompStrength.delete()

    @staticmethod
    def clearAll():
        Material.clear()
        CreepShrinkage.clear()
        CompStrength.clear()
        


# ---------------------------------  CONCRETE MATERIAL --------------------------------------------------------------

    class CONC:


        # ----------------------------------  DB MATERIAL ---------------------------------------------------

        def __init__(self,name='',standard='',db='',id:int=None,):
            if id == None: id =0  
            js =  {
                "TYPE": "CONC",
                "NAME": name,
                "DAMP_RAT": 0.05,
                "PARAM": [
                    {
                        "P_TYPE": 1,
                        "STANDARD": standard,
                        "CODE": "",
                        "DB": db,
                    }
                ]
            }
            temp = Material(js,id)
            self.ID = temp.ID
            self.DATA = js


        # ----------------------------------  USER MATERIAL ---------------------------------------------------

        class User:
            def __init__(self,name='',E=0,pois=0,den=0,mass=0,therm=0,id:int=None,):
                if id == None: id =0
                js =  {
                    "TYPE": "CONC",
                    "NAME": name,
                    "DAMP_RAT": 0.05,
                    "PARAM": [
                        {
                            "P_TYPE": 2,
                            "ELAST": E,
                            "POISN": pois,
                            "THERMAL": therm,
                            "DEN": den,
                            "MASS": mass
                        }
                    ]
                }
                temp = Material(js,id)
                self.ID = temp.ID
                self.DATA = js

    

# ---------------------------------  STEEL MATERIAL --------------------------------------------------------------

    class STEEL:

        # ----------------------------------  DB MATERIAL ---------------------------------------------------

        def __init__(self,name='',standard='',db='',id:int=None,):
            if id == None: id =0
            js =  {
                "TYPE": "STEEL",
                "NAME": name,
                "DAMP_RAT": 0.05,
                "PARAM": [
                    {
                        "P_TYPE": 1,
                        "STANDARD": standard,
                        "CODE": "",
                        "DB": db,
                    }
                ]
            }
            temp = Material(js,id)
            self.ID = temp.ID
            self.DATA = js


        # ----------------------------------  USER MATERIAL ---------------------------------------------------

        class User:
            def __init__(self,name='',E=0,pois=0,den=0,mass=0,therm=0,id:int=None,):
                if id == None: id =0
                js =  {
                    "TYPE": "STEEL",
                    "NAME": name,
                    "DAMP_RAT": 0.05,
                    "PARAM": [
                        {
                            "P_TYPE": 2,
                            "ELAST": E,
                            "POISN": pois,
                            "THERMAL": therm,
                            "DEN": den,
                            "MASS": mass
                        }
                    ]
                }
                temp = Material(js,id)
                self.ID = temp.ID
                self.DATA = js




# ---------------------------------  USER MATERIAL --------------------------------------------------------------

    class USER:

        def __init__(self,name='',E=0,pois=0,den=0,mass=0,therm=0,id:int=None,):
            if id == None: id =0
            js =  {
                "TYPE": "USER",
                "NAME": name,
                "DAMP_RAT": 0.05,
                "PARAM": [
                    {
                        "P_TYPE": 2,
                        "ELAST": E,
                        "POISN": pois,
                        "THERMAL": therm,
                        "DEN": den,
                        "MASS": mass
                    }
                ]
            }
            temp = Material(js,id)
            self.ID = temp.ID
            self.DATA = js


# ------------------------------------------ TIME DEPENDENT - CREEP and SHRINKAGE ----------------------------------------------------

class CreepShrinkage:
    mats = []
    ids = []
    def __init__(self,data,id:int=None):
        if id == None: id =0
        if CreepShrinkage.ids == []:
            count = 1
        else:
            count = max(CreepShrinkage.ids)+1
        if id == 0 or id in CreepShrinkage.ids: self.ID = count
        if id!= 0 and id not in CreepShrinkage.ids: self.ID = id

        self.DATA = data

        CreepShrinkage.mats.append(self)
        CreepShrinkage.ids.append(self.ID)

    @classmethod
    def json(cls):
        json = {"Assign":{}}
        for k in cls.mats:
            json["Assign"][k.ID]=k.DATA
        return json

    @staticmethod
    def create():
        MidasAPI("PUT","/db/TDMT",CreepShrinkage.json())

    @staticmethod
    def get():
        return MidasAPI("GET","/db/TDMT")


    @staticmethod
    def delete():
        MidasAPI("DELETE","/db/TDMT")
        CreepShrinkage.clear()

    @staticmethod
    def clear():
        CreepShrinkage.mats=[]
        CreepShrinkage.ids=[]

    @staticmethod
    def sync():
        a = CreepShrinkage.get()
        if a != {'message': ''}:
            if list(a['TDMT'].keys()) != []:
                CreepShrinkage.mats = []
                CreepShrinkage.ids=[]
                for j in a['TDMT'].keys():
                    CreepShrinkage(a['TDMT'][j], int(j))

    # ---------------------------------  IRC CnS --------------------------------------------------------------

    class IRC:
        def __init__(self,name: str, code_year: int = 2011, fck: float = 0, notional_size: float = 1,
                     relative_humidity: float = 70, age_shrinkage: int = 3, type_cement: str = 'NR', id: int = None):
            """
            IRC Creep and Shrinkage for Indian Road Congress standards. 

            Parameters:
                name (str): The name for the material property.
                code_year (int, optional): The year of the IRC code. Can be 2000 or 2011. Defaults to 2011.
                fck (float): 28-day characteristic compressive strength. 
                notional_size (float): The notional size of the member 
                relative_humidity (float): The relative humidity in percentage (40-99%). 
                age_shrinkage (int): The age of the concrete at the beginning of shrinkage in days. 
                type_cement (str, optional): The type of cement ('SL'= Slow Setting cement, 'NR'= Normal cement, 'RS'=Rapid hardening cement). Only for IRC:112-2011. Defaults to 'NR'. 
                id (int, optional): A specific ID for the material. Auto-generated if not provided.

            Examples:
                ```python
                # Create a material based on IRC:112-2011
                CreepShrinkage.IRC("IRC_M30_2011", code_year=2011, fck=30000, notional_size=1, type_cement = "RS", age_shrinkage=7)

                # Create a material based on IRC:18-2000
                CreepShrinkage.IRC("IRC_M25_2000", code_year=2000, fck=25000, notional_size=1, relative_humidity=80, age_shrinkage=3)
                ```
            """
            if id == None: id =0
            code_name = ""
            if code_year == 2011:
                code_name = "INDIA_IRC_112_2011"
            elif code_year == 2000:
                code_name = "INDIA_IRC_18_2000"
            else:
                code_name = "INDIA_IRC_112_2011"
            
            if type_cement == "SL":
                type_cement = "RS"
            elif type_cement == "RS":
                type_cement = "SL"

            js =  {
                "NAME": name,
                "CODE": code_name,
                "STR": fck,
                "HU": relative_humidity,
                "AGE": age_shrinkage,
                "MSIZE": notional_size
            }
            if code_year == 2011:
                js["CTYPE"] = type_cement

            temp = CreepShrinkage(js,id)
            self.ID = temp.ID
            self.DATA = js

    # ---------------------------------  CEB-FIP CnS --------------------------------------------------------------

    class CEB_FIP:
        def __init__(self, name: str, code_year: int = 2010, fck: float = 0, notional_size: float = 1,
                     relative_humidity: float = 70, age_shrinkage: int = 3, type_cement: str = 'RS',
                     type_of_aggregate: int = 0, id: int = None):
            """
            CEB-FIP Creep and Shrinkage for European concrete standards.

            Parameters:
                name (str): The name for the material property. 
                code_year (int, optional): Year of the CEB-FIP standard (2010, 1990, 1978). Defaults to 2010.
                fck (float): 28-day characteristic compressive strength. 
                notional_size (float): The notional size of the member. 
                relative_humidity (float): The relative humidity in percentage (40-100%). 
                age_shrinkage (int): The age of the concrete at the beginning of shrinkage in days. 
                type_cement (str, optional): The type of cement ('RS', 'NR', 'SL'). Defaults to 'RS'. 
                type_of_aggregate (int, optional): Type of aggregate, only for CEB-FIP 2010. 0: Basalt, 1: Quartzite, 2: Limestone, 3: Sandstone. Defaults to 0. 
                id (int, optional): A specific ID for the material. Auto-generated if not provided.

            Examples:
                ```python
                # Create a CEB-FIP 2010 material
                CreepShrinkage.CEB_FIP("CEB_M40", code_year=2010, fck=40000, notional_size=2, relative_humidity=65, age_shrinkage=5)

                # Create a CEB-FIP 1990 material
                CreepShrinkage.CEB_FIP("CEB_M35", code_year=1990, fck=35000, notional_size=3, relative_humidity=70, age_shrinkage=3)
                ```
            """
            if id == None: id =0
            code_name = ""
            if code_year == 2010:
                code_name = "CEB_FIP_2010"
            elif code_year == 1990:
                code_name = "CEB"
            elif code_year == 1978:
                code_name = "CEB_FIP_1978"
            else:
                code_name = "CEB_FIP_2010"

            js = {
                "NAME": name,
                "CODE": code_name,
                "STR": fck,
                "HU": relative_humidity,
                "AGE": age_shrinkage,
                "MSIZE": notional_size,
                "CTYPE": type_cement,
            }
            if code_year == 2010:
                js["TYPEOFAFFR"] = type_of_aggregate

            temp = CreepShrinkage(js, id)
            self.ID = temp.ID
            self.DATA = js

    # ---------------------------------  ACI CnS --------------------------------------------------------------

    class ACI:
        def __init__(self, name: str, fck: float = 0, relative_humidity: float = 70, age_shrinkage: int = 3,
                     vol_surface_ratio: float = 1.2, cfact_a: float = 4, cfact_b: float = 0.85,
                     curing_method: str = "MOIST", material_type: str = "CODE", cement_content: float = 24,
                     slump: float = 1.1, fine_agg_percent: float = 12, air_content: float = 13,
                     creep_coeff: float = None, shrink_strain: float = None, id: int = None):
            """
            ACI Creep and Shrinkage for American Concrete Institute standards. 

            Parameters:
                name (str): The name for the material property.
                fck (float): 28-day compressive strength . 
                relative_humidity (float): The relative humidity (40-99%). 
                age_shrinkage (int): The age of the concrete at the beginning of shrinkage in days. 
                vol_surface_ratio (float): The volume to surface area ratio. 
                cfact_a (float): Concrete compressive strength factor 'a'. 
                cfact_b (float): Concrete compressive strength factor 'b'. 
                curing_method (str, optional): Curing method ('MOIST' or 'STEAM'). Defaults to 'MOIST'. 
                material_type (str, optional): Material factored ultimate value type ('CODE' or 'USER'). Defaults to 'CODE'. 
                cement_content (float, optional): Cement content (used if material_type='CODE'). 
                slump (float, optional): Slump value (used if material_type='CODE'). 
                fine_agg_percent (float, optional): Fine aggregate percentage (used if material_type='CODE'). 
                air_content (float, optional): Air content (used if material_type='CODE'). 
                creep_coeff (float, optional): Creep coefficient (used if material_type='USER'). 
                shrink_strain (float, optional): Shrinkage strain in E-6 (used if material_type='USER'). 
                id (int, optional): A specific ID for the material. Auto-generated if not provided.

            Examples:
                ```python
                # Create an ACI material using code-based properties
                CreepShrinkage.ACI("ACI_C35_Code", fck=35000, relative_humidity=75, age_shrinkage=7, vol_surface_ratio=50)

                # Create an ACI material using user-defined ultimate values
                CreepShrinkage.ACI("ACI_C35_User", fck=35000, relative_humidity=75, age_shrinkage=7, vol_surface_ratio=50, material_type="USER", creep_coeff=2.5, shrink_strain=600)
                ```
            """
            if id == None: id =0
            js = {
                "NAME": name,
                "CODE": "ACI",
                "STR": fck,
                "HU": relative_humidity,
                "AGE": age_shrinkage,
                "VOL": vol_surface_ratio,
                "CFACTA": cfact_a,
                "CFACTB": cfact_b,
                "TYPE": material_type,
                "CMETHOD": curing_method
            }

            if material_type == "CODE":
                js.update({
                    "CEMCONTENT": cement_content,
                    "SLUMP": slump,
                    "FAPERCENT": fine_agg_percent,
                    "AIRCONTENT": air_content
                })
            elif material_type == "USER":
                js.update({
                    "CREEPCOEFF": creep_coeff if creep_coeff is not None else 1.4,
                    "SHRINKSTRAIN": shrink_strain if shrink_strain is not None else 500
                })

            temp = CreepShrinkage(js, id)
            self.ID = temp.ID
            self.DATA = js

    # ---------------------------------  AASHTO CnS --------------------------------------------------------------

    class AASHTO:
        def __init__(self, name: str, fck: float = 0, relative_humidity: float = 70, age_shrinkage: int = 3,
                     vol_surface_ratio: float = 1.2, b_expose: bool = False, id: int = None):
            """
            AASHTO Creep and Shrinkage model.

            Parameters:
                name (str): The name for the material property.
                fck (float): 28-day compressive strength.
                relative_humidity (float): The relative humidity (40-99%).
                age_shrinkage (int): The age of the concrete at the beginning of shrinkage in days.
                vol_surface_ratio (float): The volume to surface area ratio.
                b_expose (bool, optional): Expose to drying before 5 days of curing. Defaults to False.
                id (int, optional): A specific ID for the material. Auto-generated if not provided.

            Examples:
                ```python
                CreepShrinkage.AASHTO("AASHTO_M30", fck=30000, relative_humidity=80, age_shrinkage=5, vol_surface_ratio=60)
                ```
            """
            if id == None: id =0
            js = {
                "NAME": name,
                "CODE": "AASHTO",
                "STR": fck,
                "HU": relative_humidity,
                "AGE": age_shrinkage,
                "VOL": vol_surface_ratio,
                "bEXPOSE": b_expose
            }
            temp = CreepShrinkage(js, id)
            self.ID = temp.ID
            self.DATA = js

    # ---------------------------------  European CnS --------------------------------------------------------------

    class European:
        def __init__(self, name: str, fck: float = 0, relative_humidity: float = 70, age_shrinkage: int = 3,
                     notional_size: float = 1.2, type_cement: str = "Class N", t_code: int = 0, b_silica: bool = False, id: int = None):
            """
            European Creep and Shrinkage model (EN 1992). 

            Parameters:
                name (str): The name for the material property.
                fck (float): 28-day characteristic compressive strength. 
                relative_humidity (float): The relative humidity in percentage (40-99%). 
                age_shrinkage (int): The age of the concrete at the beginning of shrinkage in days. 
                notional_size (float): The notional size of the member. 
                type_cement (str, optional): Cement class ('Class S', 'Class N', 'Class R'). Defaults to 'Class N'. 
                t_code (int, optional): Type of code. 0: EN 1992-1 (General), 1: EN 1992-2 (Bridge). Defaults to 0. 
                b_silica (bool, optional): Whether silica fume is used. Only applicable when t_code is 1. Defaults to False. 
                id (int, optional): A specific ID for the material. Auto-generated if not provided.

            Examples:
                ```python
                # EN 1992-1 General Structure
                CreepShrinkage.European("Euro_General_C30", fck=30000, relative_humidity=75, age_shrinkage=7, notional_size=2)

                # EN 1992-2 Concrete Bridge with Silica Fume
                CreepShrinkage.European("Euro_Bridge_C40", fck=40000, relative_humidity=70, age_shrinkage=5, notional_size=2, t_code=1, b_silica=True)
                ```
            """
            if id == None: id =0
            js = {
                "NAME": name,
                "CODE": "EUROPEAN",
                "STR": fck,
                "HU": relative_humidity,
                "AGE": age_shrinkage,
                "MSIZE": notional_size,
                "CTYPE": type_cement,
                "TCODE": t_code,
            }
            if t_code == 1:
                js["bSILICA"] = b_silica

            temp = CreepShrinkage(js, id)
            self.ID = temp.ID
            self.DATA = js

    # ---------------------------------  Russian CnS --------------------------------------------------------------
    class Russian:
        def __init__(self, name: str, fck: float, relative_humidity: float, module_exposed_surface: float,
                     age_concrete: int, water_content: float, max_aggregate_size: float, air_content: float,
                     specific_cement_paste_content: float, curing_method: int = 0,cement_type=1, fast_accumulating_creep: bool = False,
                     concrete_type: int = 0, id: int = None):
            """
            Russian Creep and Shrinkage model. 

            Parameters:
                name (str): The name for the material property. 
                fck (float): 28-day compressive strength. 
                relative_humidity (float): The relative humidity in percentage. 
                module_exposed_surface (float): The module of an exposed surface. 
                age_concrete (int): The age of the concrete in days. 
                water_content (float): The water content . 
                max_aggregate_size (float): Maximum aggregate size. 
                air_content (float): Air content. 
                specific_cement_paste_content (float): Specific content of the cement paste. 
                curing_method (int, optional): Curing method. 0: Natural, 1: Steam. Defaults to 0.
                cement_type(int ,optional) : Cement Type. 0: Normal (Default), 1: Fast-hardened , 2: Slag ,3: Pozzolan
                fast_accumulating_creep (bool, optional): Whether to consider fast-accumulating creep. Defaults to False. 
                concrete_type (int, optional): Type of concrete. 0: Heavy, 1: Fine-grained. Defaults to 0. 
                id (int, optional): A specific ID for the material. Auto-generated if not provided.

            Examples:
                ```python
                # Standard Russian model with natural curing
                CreepShrinkage.Russian("RU_Heavy_C30", fck=30000, relative_humidity=70, module_exposed_surface=10, age_concrete=14, water_content=180, max_aggregate_size=0.02, air_content=30, specific_cement_paste_content=0.25,cement_type=2)
                ```
            """
            if id == None: id =0
            js = {
                "NAME": name,
                "CODE": "RUSSIAN",
                "STR": fck,
                "HU": relative_humidity,
                "M": module_exposed_surface,
                "AGE": age_concrete,
                "CMETH": curing_method,
                "iCTYPE": cement_type,
                "CREEP": fast_accumulating_creep,
                "CONCT": concrete_type,
                "W": water_content,
                "MAXS": max_aggregate_size,
                "A": air_content,
                "PZ": specific_cement_paste_content
            }
            temp = CreepShrinkage(js, id)
            self.ID = temp.ID
            self.DATA = js

    # ---------------------------------  AS & NZ CnS -------------------------------------------------
    class AS_NZ:
        def __init__(self, name: str, standard: str, fck: float, concrete_age: int,
                    hypothetical_thickness: float, drying_shrinkage_type: int = 0,
                    user_defined_shrinkage_strain: float = 0, humidity_factor: float = 0.72,
                    exposure_environment: int = 0, id: int = None):
            """
            Australian & New Zealand Standards Creep and Shrinkage model.

            Parameters:
                name (str): The name for the material property.
                standard (str): The standard code. Valid codes are 'AS_5100_5_2017', 'AS_5100_5_2016', 'AS_RTA_5100_5_2011', 'AS_3600_2009', 'NEWZEALAND'.
                fck (float): 28-day compressive strength
                concrete_age (int): The age of the concrete in days
                hypothetical_thickness (float): The hypothetical thickness of the member
                drying_shrinkage_type (int, optional): Type of drying basic shrinkage strain,The values depend on the chosen standard.
                    - For AS Standards ('AS_5100_5_2017', 'AS_5100_5_2016', etc.):
                    - 0: 800.0 (Sydney, Brisbane)
                    - 1: 900.0 (Melbourne)
                    - 2: 1000.0 (Elsewhere)
                    - 3: User Defined Value
                - For NZ Bridge Standard ('NEWZEALAND'):
                    - 0: 1500 (Hastings, Palmerston North, etc.)
                    - 1: 1460 (Nelson)
                    - 2: 1315 (Kaitaia, Tauranga)
                    - 3: 1080 (New Plymouth, Taranaki)
                    - 4: 1000 (Whangarei, Auckland Hunua, Hamilton)
                    - 5: 990 (Auckland)
                    - 6: 950 (Christchurch, Timaru, etc.)
                    - 7: 775 (Westport, Queenstown, etc.)
                    - 8: 735 (Dunedin)
                    - 9: 570 (Waiapu)
                    - 10: User Defined Value
                user_defined_shrinkage_strain (float, optional): The user-defined drying basic shrinkage strain, This value is ONLY used when `drying_shrinkage_type` is set to a user-defined option (3 for AS, 10 for NZ).
                humidity_factor (float, optional): Relative humidity thickness. This is only for the NZ Bridge standard and has a range of 0.20 to 0.72. Defaults to 0.72.
                exposure_environment (int, optional): Exposure environment classification ("EXPOSURE") for AS standards. Defaults to 0.
                id (int, optional): A specific ID for the material. Auto-generated if not provided.

            Examples:
                ```python
                # AS 5100.5-2017 
                CreepShrinkage.AS_NZ("AS_Melbourne_C40", standard="AS_5100_5_2017", fck=40000, hypothetical_thickness=0.3, concrete_age=7, drying_shrinkage_type=1)

                # NZ Bridge standard 
                CreepShrinkage.AS_NZ("NZ_Bridge_Custom", standard="NEWZEALAND", fck=35000, hypothetical_thickness=0.25, concrete_age=14, drying_shrinkage_type=10, user_defined_shrinkage_strain=955.5)
                ```
            """
            if id == None: id =0
            js = {
                "NAME": name,
                "CODE": standard,
                "STR": fck,
                "THIK": hypothetical_thickness,
                "AGE": concrete_age,
                "iEPS_DRY": drying_shrinkage_type,
            }

            # Internal maps for predefined shrinkage values from the manual
            as_strain_map = {0: 800.0, 1: 900.0, 2: 1000.0}
            nz_strain_map = {0: 1500.0, 1: 1460.0, 2: 1315.0, 3: 1080.0, 4: 1000.0, 5: 990.0, 6: 950.0, 7: 775.0, 8: 735.0, 9: 570.0}

            eps_dry_value = None
            is_as_code = standard != "NEWZEALAND"
            is_nz_code = standard == "NEWZEALAND"

            # Check for user-defined cases first
            if is_as_code and drying_shrinkage_type == 3:
                eps_dry_value = user_defined_shrinkage_strain
            elif is_nz_code and drying_shrinkage_type == 10:
                eps_dry_value = user_defined_shrinkage_strain
            # Otherwise, look up the predefined value from the appropriate map
            elif is_as_code:
                eps_dry_value = as_strain_map.get(drying_shrinkage_type)
            elif is_nz_code:
                eps_dry_value = nz_strain_map.get(drying_shrinkage_type)

            js["EPS_DRY"] = eps_dry_value

            # Add parameters specific to the standard
            if is_nz_code:
                js["FS"] = humidity_factor
            else: # Assumes all other codes are AS standards
                js["EXPOSURE"] = exposure_environment

            temp = CreepShrinkage(js, id)
            self.ID = temp.ID
            self.DATA = js

    # ---------------------------------  Chinese Standard CnS ----------------------------------------------------

    class Chinese:
        def __init__(self, name: str, standard: str, fck: float, relative_humidity: float,
                     concrete_age: int, notional_size: float, humidity_type: str = "RH",
                     cement_coeff: float = 5, fly_ash_amount: float = 20, id: int = None):
            """
            Chinese Standards Creep and Shrinkage model.

            Parameters:
                name (str): The name for the material property. 
                standard (str): The Chinese standard code ('CHINESE', 'JTG', 'CHINA_JTG3362_2018'). 
                fck (float): 28-day compressive strength. 
                relative_humidity (float): The relative humidity in percentage. 
                concrete_age (int): The age of the concrete in days. 
                notional_size (float): The notional size of the member. 
                humidity_type (str, optional): Type of relative humidity ('CU' for Curing Underwater, 'RH' for Relative Humidity). Defaults to 'RH'. 
                cement_coeff (float, optional): Cement type coefficient (for JTG codes). Defaults to 5. 
                fly_ash_amount (float, optional): Amount of added fly ash (for JTG3362-2018). Range 0-30. Defaults to 20. 
                id (int, optional): A specific ID for the material. Auto-generated if not provided.

            Examples:
                ```python
                # General Chinese Standard
                CreepShrinkage.Chinese("Chinese_C30", standard="CHINESE", fck=30000, relative_humidity=75, concrete_age=14, notional_size=2)

                # JTG D62-2004 Standard
                CreepShrinkage.Chinese("JTG_D62_C40", standard="JTG", fck=40000, relative_humidity=80, concrete_age=7, notional_size=250, cement_coeff=5)
                ```
            """
            if id == None: id =0
            js = {
                "NAME": name,
                "CODE": standard,
                "STR": fck,
                "HU": relative_humidity,
                "AGE": concrete_age,
                "MSIZE": notional_size,
                "HTYPE": humidity_type
            }
            if "JTG" in standard:
                js["BSC"] = cement_coeff
            if standard == "CHINA_JTG3362_2018":
                js["FLYASH"] = fly_ash_amount

            temp = CreepShrinkage(js, id)
            self.ID = temp.ID
            self.DATA = js

    # ---------------------------------  Korean Standards CnS ----------------------------------------------------

    class Korean:
        def __init__(self, name: str, standard: str, fck: float, relative_humidity: float,
                     concrete_age: int, notional_size: float, cement_type: str = "NR",
                     density: float = 240, id: int = None):
            """
            Korean Standards Creep and Shrinkage model.

            Parameters:
                name (str): The name for the material property. 
                standard (str): The Korean standard code ('KDS_2016', 'KSI_USD12', 'KSCE_2010', 'KS'). 
                fck (float): 28-day compressive strength. 
                relative_humidity (float): The relative humidity in percentage. 
                concrete_age (int): The age of the concrete in days. 
                notional_size (float): The notional size of the member. 
                cement_type (str, optional): The type of cement ('SL', 'NR', 'RS'). Defaults to 'NR'. 
                density (float, optional): Weight density  (Only for KDS-2016). Defaults to 240. 
                id (int, optional): A specific ID for the material. Auto-generated if not provided.

            Examples:
                ```python
                # KDS-2016 Standard
                CreepShrinkage.Korean("KDS_C35", standard="KDS_2016", fck=35000, relative_humidity=65, concrete_age=10, notional_size=2, density=2400)

                # Korea Standard (KS)
                CreepShrinkage.Korean("KS_C30", standard="KS", fck=30000, relative_humidity=70, concrete_age=14, notional_size=2)
                ```
            """
            if id == None: id =0
            js = {
                "NAME": name,
                "CODE": standard,
                "STR": fck,
                "HU": relative_humidity,
                "AGE": concrete_age,
                "MSIZE": notional_size,
                "CTYPE": cement_type
            }
            if standard == "KDS_2016":
                js["DENSITY"] = density

            temp = CreepShrinkage(js, id)
            self.ID = temp.ID
            self.DATA = js

    # ---------------------------------  PCA CnS -----------------------------------------------------------------

    class PCA:
        def __init__(self, name: str, fck: float, relative_humidity: float, ultimate_creep_strain: float,
                     vol_surface_ratio: float, reinforcement_ratio: float, steel_elasticity_modulus: float,
                     ultimate_shrinkage_strain: float, id: int = None):
            """
            PCA Creep and Shrinkage model.

            Parameters:
                name (str): The name for the material property. 
                fc (float): Compressive strength
                relative_humidity (float): The relative humidity in percentage. (Range : 40-99) 
                ultimate_creep_strain (float): Ultimate creep strain. (Range : 3-5) 
                vol_surface_ratio (float): The volume to surface area ratio. 
                reinforcement_ratio (float): Reinforcement ratio of the cross section. 
                steel_elasticity_modulus (float): Modulus of elasticity of steel. 
                ultimate_shrinkage_strain (float): Ultimate shrinkage strain. (Range : 500-800) 
                id (int, optional): A specific ID for the material. Auto-generated if not provided.

            Examples:
                ```python
                CreepShrinkage.PCA("PCA_Material", fck=50000, relative_humidity=70, ultimate_creep_strain=4, vol_surface_ratio=1.2, reinforcement_ratio=20, steel_elasticity_modulus=2e8, ultimate_shrinkage_strain=780)
                ```
            """
            if id == None: id =0
            js = {
                "NAME": name,
                "CODE": "PCA",
                "STR": fck,
                "HU": relative_humidity,
                "UCS": ultimate_creep_strain,
                "VOL": vol_surface_ratio,
                "RR": reinforcement_ratio,
                "MOD": steel_elasticity_modulus,
                "USS": ultimate_shrinkage_strain
            }
            temp = CreepShrinkage(js, id)
            self.ID = temp.ID
            self.DATA = js

    # ---------------------------------  Japan CnS ---------------------------------------------------------------

    class Japan:
        def __init__(self, name: str, standard: str, relative_humidity: float, concrete_age: int,
                     vol_surface_ratio: float, cement_content: float, water_content: float, fck: float = 30000,
                     impact_factor: float = 1, age_of_solidification: int = 5, alpha_factor: int = 11,
                     autogenous_shrinkage: bool = True, gamma_factor: int = 1, a_factor: float = 0.1,
                     b_factor: float = 0.7, general_shrinkage: bool = True, id: int = None):
            """
            Japan Creep and Shrinkage model (JSCE). 

            Parameters:
                name (str): The name for the material property. 
                standard (str): The Japanese standard ('JSCE_12', 'JSCE_07', 'JSCE'). 
                relative_humidity (float): The relative humidity in percentage. 
                concrete_age (int): The age of the concrete in days. 
                vol_surface_ratio (float): The volume to surface area ratio. 
                cement_content (float): Cement content. 
                water_content (float): Water content. 
                fck (float, optional): Compressive strength (not for JSCE). Defaults to 30000. 
                impact_factor (float, optional): Impact factor by cement type (JSCE 2012 only). Defaults to 1. 
                age_of_solidification (int, optional): Age at beginning of solidification (JSCE 2012 only). Defaults to 5. 
                alpha_factor (int, optional): Alpha-factor by cement type (JSCE 2007 only). Defaults to 11. 
                autogenous_shrinkage (bool, optional): Autogenous shrinkage option (JSCE 2007 only). Defaults to True. 
                gamma_factor (int, optional): Gamma-factor (JSCE 2007 only). Defaults to 1. 
                a_factor (float, optional): a-factor (JSCE 2007 only). Defaults to 0.1. 
                b_factor (float, optional): b-factor (JSCE 2007 only). Defaults to 0.7. 
                general_shrinkage (bool, optional): General shrinkage option (JSCE 2007 only). Defaults to True. 
                id (int, optional): A specific ID for the material. Auto-generated if not provided.

            Examples:
                ```python
                # JSCE 2012
                CreepShrinkage.Japan("JSCE12_mat", "JSCE_12", 70, 3, 0.2, 30, 20, fck=30000)
                # JSCE 2007
                CreepShrinkage.Japan("JSCE07_mat", "JSCE_07", 70, 3, 0.2, 30, 20, fck=30000, alpha_factor=15)
                ```
            """
            if id == None: id =0
            js = {
                "NAME": name,
                "CODE": standard,
                "HU": relative_humidity,
                "AGE": concrete_age,
                "VOL": vol_surface_ratio,
                "CEMCONTENT": cement_content,
                "WATERCONTENT": water_content
            }
            if standard != "JSCE":
                js["STR"] = fck
            if standard == "JSCE_12":
                js["IPFACT"] = impact_factor
                js["AGESOL"] = age_of_solidification
            if standard == "JSCE_07":
                js["ALPHAFACT"] = alpha_factor
                js["bAUTO"] = autogenous_shrinkage
                if autogenous_shrinkage:
                    js["GAMMAFACT"] = gamma_factor
                    js["AFACT"] = a_factor
                    js["BFACT"] = b_factor
                js["bGEN"] = general_shrinkage

            temp = CreepShrinkage(js, id)
            self.ID = temp.ID
            self.DATA = js

    # ---------------------------------  Japanese Standard CnS ---------------------------------------------
    class JapaneseStandard:
        def __init__(self, name: str, fck: float, relative_humidity: float, concrete_age: int, notional_size: float,
                     calculation_method: str = "JSCE", humidity_type: str = "RH", cement_type: str = "NC",
                     environmental_coeff: int = 1, id: int = None):
            """
            Japanese Standard Creep and Shrinkage model. 

            Parameters:
                name (str): The name for the material property. 
                fck (float): Compressive strength 
                relative_humidity (float): The relative humidity in percentage (40-90%). 
                concrete_age (int): The age of the concrete in days. 
                notional_size (float): The notional size of the member. 
                calculation_method (str, optional): Calculation method for E ('JSCE' or 'AIJ'). Defaults to 'JSCE'. 
                humidity_type (str, optional): Relative humidity type ('RH' or 'CU'). Defaults to 'RH'. 
                cement_type (str, optional): Type of cement ('RH', 'NC'). Defaults to 'NC'. 
                environmental_coeff (int, optional): Environmental coefficient. Defaults to 1. 
                id (int, optional): A specific ID for the material. Auto-generated if not provided.

            Examples:
                ```python
                CreepShrinkage.JapaneseStandard("JapanStd_C30", fck=30000, relative_humidity=70, concrete_age=3, notional_size=1.2)
                ```
            """
            if id == None: id =0
            js = {
                "NAME": name,
                "CODE": "JAPAN",
                "STR": fck,
                "HU": relative_humidity,
                "HTYPE": humidity_type,
                "AGE": concrete_age,
                "MSIZE": notional_size,
                "CTYPE": cement_type,
                "CM": calculation_method,
                "LAMBDA": environmental_coeff
            }
            temp = CreepShrinkage(js, id)
            self.ID = temp.ID
            self.DATA = js

    # ---------------------------------  User Defined CnS ----------------------------------------------------

    class UserDefined:
        def __init__(self, name: str, shrinkage_func_name: str, creep_func_name: str, creep_age: int, id: int = None):
            """
            User Defined Creep and Shrinkage model.

            Parameters:
                name (str): The name for the material property. 
                shrinkage_func_name (str): The name of the user-defined shrinkage strain function. 
                creep_func_name (str): The name of the user-defined creep function. 
                creep_age (int): Concrete age for the creep function. 
                id (int, optional): A specific ID for the material. Auto-generated if not provided.
            """
            if id == None: id =0
            js = {
                "NAME": name,
                "CODE": "USER_DEFINED",
                "SSFNAME": shrinkage_func_name,
                "vCREEP_AGE": [
                    {
                        "NAME": creep_func_name,
                        "AGE": creep_age
                    }
                ]
            }
            temp = CreepShrinkage(js, id)
            self.ID = temp.ID
            self.DATA = js

#------------------------------------------ TIME DEPENDENT - COMPRESSIVE STRENGTH ----------------------------------------------------



class CompStrength:
    mats = []
    ids = []
    def __init__(self,data,id=None):
        if id == None: id =0
        if CompStrength.ids == []: 
            count = 1
        else:
            count = max(CompStrength.ids)+1
        if id == 0 or id in CompStrength.ids: self.ID = count
        if id!= 0 and id not in CompStrength.ids: self.ID = id

        self.DATA = data

        CompStrength.mats.append(self)
        CompStrength.ids.append(self.ID)
    
    @classmethod
    def json(cls):
        json = {"Assign":{}}
        for k in cls.mats:
            json["Assign"][k.ID]=k.DATA
        return json
    
    @staticmethod
    def create():
        MidasAPI("PUT","/db/TDME",CompStrength.json())
        
    @staticmethod
    def get():
        return MidasAPI("GET","/db/TDME")
    
    @staticmethod
    def delete():
        MidasAPI("DELETE","/db/TDME")
        CompStrength.clear()

    @staticmethod
    def clear():
        CompStrength.mats=[]
        CompStrength.ids=[]

    @staticmethod
    def sync():
        a = CompStrength.get()
        if a != {'message': ''}:
            if list(a['TDME'].keys()) != []:
                CompStrength.mats = []
                CompStrength.ids=[]
                for j in a['TDME'].keys():
                    CompStrength(a['TDME'][j], int(j))


    # ---------------------------------  IRC Compressive Strength --------------------------------------------------------------

    class IRC:
        def __init__(self, name: str, code_year: int = 2020,
                     fck_delta: float = 0, cement_type: int = 1,
                     aggregate_type: int = 0, id: int = None):
            """
            IRC Compressive Strength for Indian Road Congress standards.

            Parameters:
                name (str): Name 
                code_year (int, optional): Year of the IRC standard. Can be 2020, 2011, or 2000. Defaults to 2020.
                fck_delta (float): 28-day characteristic compressive strength
                cement_type (int, optional): Type of cement used. 
                    • 1: Slow setting (default)
                    • 2: Normal 
                    • 3: Rapid hardening
                aggregate_type (int, optional): Type of aggregate used (for IRC:112 only).
                    • 0: Basalt, dense limestone (default)
                    • 1: Quartzite 
                    • 2: Limestone 
                    • 3: Sandstone
                id (int, optional): Unique identifier. Auto-generated if not specified.

            Examples:
                ```python
                # IRC 112-2020 with normal cement
                CompStrength.IRC("C30_IRC2020", code_year=2020, fck_delta=30000, cement_type=2)

                # IRC 18-2000 standard
                CompStrength.IRC("C25_IRC2000", code_year=2000, fck_delta=25000)

                # IRC 112-2011 with rapid hardening cement and quartzite aggregate
                CompStrength.IRC("C40_IRC2011", code_year=2011, fck_delta=40000, cement_type=3, aggregate_type=1)
                ```
            """
            if id == None: id =0
            # Determine the code name string based on the integer year
            if code_year == 2011:
                code_name = "INDIA(IRC:112-2011)"
            elif code_year == 2000:
                code_name = "INDIA(IRC:18-2000)"
            else: # Default to 2020
                code_name = "INDIA(IRC:112-2020)"

            js = {
                "NAME": name,
                "TYPE": "CODE",
                "CODENAME": code_name,
                "STRENGTH": fck_delta
            }

            # Add cement and aggregate types for IRC:112 standards
            if code_year in [2020, 2011]:
                js["iCTYPE"] = cement_type
                js["nAGGRE"] = aggregate_type

            temp = CompStrength(js, id)
            self.ID = temp.ID
            self.DATA = js


    # ---------------------------------  ACI Compressive Strength --------------------------------------------------------------

    class ACI:
        def __init__(self, name: str, fck: float = 0, factor_a: float = 1, 
                     factor_b: float = 2, id: int = None):
            """
            ACI Compressive Strength for American Concrete Institute standards.
            
            Parameters:
                name: Name
                fck: Compression strength (number) - 28-day compressive strength
                factor_a: Factor a (number) - ACI model parameter A (default 1)
                factor_b: Factor b (number) - ACI model parameter B (default 2)
                id: Unique identifier (integer) - Auto-generated if not specified
            
            Examples:
                ```python
                # Standard ACI material
                CompStrength.ACI("C30_ACI", 30000, 1, 2)
                
                # Custom factors
                CompStrength.ACI("C25_ACI_Custom", 25000, 1.2, 1.8)
                ```
            """
            if id == None: id =0
            js = {
                "NAME": name,
                "TYPE": "CODE",
                "CODENAME": "ACI",
                "STRENGTH": fck,
                "A": factor_a,
                "B": factor_b
            }
            temp = CompStrength(js, id)
            self.ID = temp.ID
            self.DATA = js


    # ---------------------------------  CEB-FIP Compressive Strength --------------------------------------------------------------

    class CEB_FIP:
        def __init__(self, name: str, code_year: int = 2010, fck: float = 0, 
                     cement_type: int = 1, aggregate_type: int = 0, id: int = None):
            """
            CEB-FIP Compressive Strength for European concrete standards.
            
            Parameters:
                name: Name
                code_year: Code year (integer) - Year of the CEB-FIP standard
                    • 1978 - CEB-FIP(1978)
                    • 1990 - CEB-FIP(1990) 
                    • 2010 - CEB-FIP(2010) (default)
                fck: Compression strength (number) - 28-day compressive strength
                cement_type: Cement type (integer) - Type of cement (for 1990 and 2010)
                    • 1 - RS: 0.2 / 42.5 R, 52.5 N, 52.5 R: 0.20 (default)
                    • 2 - N, R: 0.25 / 32.5 R, 42.5 N: 0.25
                    • 3 - SL: 0.38 / 32.5 N: 0.38
                aggregate_type: Aggregate type (integer) - Type of aggregate (for 2010 only)
                    • 0 - Basalt, dense limestone aggregates (1.2): 0 (default)
                    • 1 - Quartzite aggregates: 1
                    • 2 - Limestone aggregates: 2
                    • 3 - Sandstone aggregates: 3
                id: Unique identifier (integer) - Auto-generated if not specified
            
            Examples:
                ```python
                # CEB-FIP 2010 with normal cement and basalt aggregate
                CompStrength.CEB_FIP("C30_CEBFIP2010", 2010, 30000, 1, 0)
                
                # CEB-FIP 1990 with slow setting cement
                CompStrength.CEB_FIP("C25_CEBFIP1990", 1990, 25000, 3)
                
                # CEB-FIP 1978 (no cement/aggregate type)
                CompStrength.CEB_FIP("C40_CEBFIP1978", 1978, 40000)
                ```
            """
            if id == None: id =0
            # Determine code name based on year
            if code_year == 1978:
                code_name = "CEB-FIP(1978)"
            elif code_year == 1990:
                code_name = "CEB-FIP(1990)"
            else:  # Default to 2010
                code_name = "CEB-FIP(2010)"
            
            js = {
                "NAME": name,
                "TYPE": "CODE",
                "CODENAME": code_name,
                "STRENGTH": fck
            }
            
            # Add cement type for 1990 and 2010
            if code_year in [1990, 2010]:
                js["iCTYPE"] = cement_type
                
            # Add aggregate type for 2010 only
            if code_year == 2010:
                js["nAGGRE"] = aggregate_type
                
            temp = CompStrength(js, id)
            self.ID = temp.ID
            self.DATA = js


    # ---------------------------------  Ohzagi Compressive Strength --------------------------------------------------------------

    class Ohzagi:
        def __init__(self, name: str, fck: float = 0, cement_type: int = 2, id: int = None):
            """
            Ohzagi Compressive Strength model.
            
            Parameters:
                name: Name
                fck: Compression strength (number) - 28-day compressive strength
                cement_type: Cement type (integer) - Type of cement used
                    • 1 - RS
                    • 2 - N, R (default)
                    • 3 - SL
                    • 4 - Fly-ash
                id: Unique identifier (integer) - Auto-generated if not specified
            
            Examples:
                ```python
                # Standard Ohzagi material with N,R cement
                CompStrength.Ohzagi("C30_Ohzagi", 30000, 2)
                
                # Fly-ash cement type
                CompStrength.Ohzagi("C25_Ohzagi_FA", 25000, 4)
                ```
            """
            if id == None: id =0
            js = {
                "NAME": name,
                "TYPE": "CODE",
                "CODENAME": "Ohzagi",
                "STRENGTH": fck,
                "iCTYPE": cement_type
            }
            temp = CompStrength(js, id)
            self.ID = temp.ID
            self.DATA = js


    # ---------------------------------  European Compressive Strength --------------------------------------------------------------

    class European:
        def __init__(self, name: str, fck: float = 0, cement_type: int = 2, id: int = None):
            """
            European Compressive Strength model.
            
            Parameters:
                name: Name
                fck: Compression strength (number) - 28-day compressive strength
                cement_type: Cement type (integer) - Cement class type
                    • 1 - Class R: 0.20
                    • 2 - Class N: 0.25 (default)
                    • 3 - Class S: 0.38
                id: Unique identifier (integer) - Auto-generated if not specified
            
            Examples:
                ```python
                # European standard with Class N cement
                CompStrength.European("C30_Euro", 30000, 2)
                
                # High early strength with Class R cement
                CompStrength.European("C40_Euro_R", 40000, 1)
                ```
            """
            if id == None: id =0
            js = {
                "NAME": name,
                "TYPE": "CODE",
                "CODENAME": "EUROPEAN",
                "STRENGTH": fck,
                "iCTYPE": cement_type
            }
            temp = CompStrength(js, id)
            self.ID = temp.ID
            self.DATA = js


    # ---------------------------------  Russian Compressive Strength --------------------------------------------------------------

    class Russian:
        def __init__(self, name: str, fck: float = 0, cement_type: int = 1, 
                     curing_method: int = 1, concrete_type: int = 1, 
                     max_aggregate_size: float = 0.02, specific_cement_content: float = 0.25, 
                     id: int = None):
            """
            Russian Compressive Strength model.
            
            Parameters:
                name: Name
                fck: Compression strength (number) - 28-day compressive strength
                cement_type: Cement type (integer) - Type of cement
                    • 1 - Normal (default)
                    • 2 - Fast-hardened
                    • 3 - Slag
                    • 4 - Pozzolan
                curing_method: Curing method (integer) - Method of curing
                    • 0 - Natural cure: 0
                    • 1 - Steam cure: 1 (default)
                concrete_type: Concrete type (integer) - Type of concrete
                    • 0 - Heavy concrete: 0
                    • 1 - Fine-grained concrete: 1 (default)
                max_aggregate_size: Maximum aggregate size (number) - Maximum size in meters (default 0.02)
                specific_cement_content: Specific content of cement paste (number) - Content ratio (default 0)
                id: Unique identifier (integer) - Auto-generated if not specified
            
            Examples:
                ```python
                # Standard Russian concrete
                CompStrength.Russian("C30_RU", 30000, 1, 1, 1, 0.02, 0)
                
                # Fast-hardened cement with natural curing
                CompStrength.Russian("C25_RU_FH", 25000, 2, 0, 0, 0.025, 0)
                ```
            """
            if id == None: id =0
            js = {
                "NAME": name,
                "TYPE": "CODE",
                "CODENAME": "RUSSIAN",
                "STRENGTH": fck,
                "iCTYPE": cement_type,
                "CMETH": curing_method,
                "CTYPE": concrete_type,
                "MAXS": max_aggregate_size,
                "PZ": specific_cement_content
            }
            temp = CompStrength(js, id)
            self.ID = temp.ID
            self.DATA = js


    # ---------------------------------  Australian Standards Compressive Strength --------------------------------------------------------------
#add EXPOSURE
    class AS:
        def __init__(self, name: str, standard: str = "AS5100.5-2017", fck: float = 0, id: int = None):
            """
            Australian Standards Compressive Strength model.
            
            Parameters:
                name: Name
                standard: Standard code (string) - Australian standard specification
                    • "AS5100.5-2017" - AS 5100.5-2017 (default)
                    • "AS5100.5-2016" - AS 5100.5-2016
                    • "AS/RTA5100.5-2011" - AS/RTA 5100.5-2011
                    • "AS3600-2009" - AS 3600-2009
                fck: Compression strength (number) - 28-day compressive strength
                id: Unique identifier (integer) - Auto-generated if not specified
            
            Examples:
                ```python
                # AS 5100.5-2017 standard
                CompStrength.AS("C30_AS2017", "AS 5100.5-2017", 30000)
                
                # AS 3600-2009 standard
                CompStrength.AS("C25_AS3600", "AS 3600-2009", 25000)
                ```
            """
            if id == None: id =0
            js = {
                "NAME": name,
                "TYPE": "CODE",
                "CODENAME": standard,
                "STRENGTH": fck
            }
            temp = CompStrength(js, id)
            self.ID = temp.ID
            self.DATA = js


    # ---------------------------------  Gilbert and Ranzi Compressive Strength --------------------------------------------------------------

    class GilbertRanzi:
        def __init__(self, name: str, fck: float = 0, cement_type: int = 1, 
                     density: float = 230, id: int = None):
            """
            Gilbert and Ranzi Compressive Strength model.
            
            Parameters:
                name: Name
                fck: Compression strength (number) - 28-day compressive strength
                cement_type: Cement type (integer) - Type of cement
                    • 1 - Ordinary Portland cement: 0.38 (default)
                    • 2 - High early strength cement: 0.25
                density: Weight density (number) - Density in kg/m³ (default 230)
                id: Unique identifier (integer) - Auto-generated if not specified
            
            Examples:
                ```python
                # Standard Gilbert-Ranzi model
                CompStrength.GilbertRanzi("C30_GR", 30000, 1, 2400)
                
                # High early strength cement
                CompStrength.GilbertRanzi("C40_GR_HES", 40000, 2, 2450)
                ```
            """
            if id == None: id =0
            js = {
                "NAME": name,
                "TYPE": "CODE",
                "CODENAME": "GILBERT AND RANZI",
                "STRENGTH": fck,
                "iCTYPE": cement_type,
                "DENSITY": density
            }
            temp = CompStrength(js, id)
            self.ID = temp.ID
            self.DATA = js


    # ---------------------------------  Japan Hydration Compressive Strength --------------------------------------------------------------

    class JapanHydration:
        def __init__(self, name: str, fck: float = 0, cement_type: int = 1, 
                     use_concrete_data: bool = True, tensile_strength_factor: float = 3,
                     factor_a: float = 4.5, factor_b: float = 0.95, factor_d: float = 1.11,
                     id: int = None):
            """
            Japan Hydration Compressive Strength model.
            
            Parameters:
                name: Name
                fck: Compression strength (number) - 28-day compressive strength
                cement_type: Cement type (integer) - Type of cement
                    • 0 - Normal Portland cement
                    • 1 - Moderate Portland cement (default)
                    • 2 - High-early-strength Portland cement
                use_concrete_data: Use concrete data option (boolean) - Enable concrete data option
                    • True - Use concrete data (default)
                    • False - Use custom factors
                tensile_strength_factor: Tensile strength factor (number) - Factor for tensile strength (default 3)
                factor_a: Factor a (number) - Model parameter A (default 4.5, used when use_concrete_data=False)
                factor_b: Factor b (number) - Model parameter B (default 0.95, used when use_concrete_data=False)
                factor_d: Factor d (number) - Model parameter D (default 1.11, used when use_concrete_data=False)
                id: Unique identifier (integer) - Auto-generated if not specified
            
            Examples:
                ```python
                # Standard Japan hydration with concrete data
                CompStrength.JapanHydration("C30_JH", 30000, 1, True, 3)
                
                # Custom factors without concrete data
                CompStrength.JapanHydration("C25_JH_Custom", 25000, 0, False, 3, 4.0, 0.9, 1.0)
                ```
            """
            if id == None: id =0
            js = {
                "NAME": name,
                "TYPE": "CODE",
                "CODENAME": "Japan(hydration)",
                "STRENGTH": fck,
                "iCTYPE": cement_type,
                "bUSE": use_concrete_data,
                "TENS_STRN_FACTOR": tensile_strength_factor
            }
            
            # Add custom factors if not using concrete data
            if not use_concrete_data:
                js.update({
                    "A": factor_a,
                    "B": factor_b,
                    "D": factor_d
                })
                
            temp = CompStrength(js, id)
            self.ID = temp.ID
            self.DATA = js


    # ---------------------------------  Japan Elastic Compressive Strength --------------------------------------------------------------

    class JapanElastic:
        def __init__(self, name: str, fck: float = 0, elastic_cement_type: int = 0, id: int = None):
            """
            Japan Elastic Compressive Strength model.
            
            Parameters:
                name: Name
                fck: Compression strength (number) - 28-day compressive strength
                elastic_cement_type: Elastic cement type (integer) - Type of cement for elastic model
                    • 0 - Normal type: 0 (default)
                    • 1 - Rapid type: 1
                id: Unique identifier (integer) - Auto-generated if not specified
            
            Examples:
                ```python
                # Normal type cement
                CompStrength.JapanElastic("C30_JE", 30000, 0)
                
                # Rapid type cement
                CompStrength.JapanElastic("C40_JE_Rapid", 40000, 1)
                ```
            """
            if id == None: id =0
            js = {
                "NAME": name,
                "TYPE": "CODE",
                "CODENAME": "Japan(elastic)",
                "STRENGTH": fck,
                "iECTYPE": elastic_cement_type
            }
            temp = CompStrength(js, id)
            self.ID = temp.ID
            self.DATA = js


    # ---------------------------------  KDS-2016 Compressive Strength --------------------------------------------------------------

    class KDS:
        def __init__(self, name: str, fck: float = 0, cement_type: int = 1, 
                     density: float = 230, id: int = None):
            """
            KDS-2016 Compressive Strength model.
            
            Parameters:
                name: Name
                fck: Compression strength (number) - 28-day compressive strength 
                cement_type: Cement type (integer) - Type of cement
                    • 1 - N,R moist cured: 0.35 (default)
                    • 2 - N,R steam cured: 0.15
                    • 3 - RS moist cured: 0.25
                    • 4 - RS steam cured: 0.12
                    • 5 - SL: 0.40
                density: Weight density (number) - Density in kg/m³ (default 230)
                id: Unique identifier (integer) - Auto-generated if not specified
            
            Examples:
                ```python
                # Standard KDS with N,R moist cured cement
                CompStrength.KDS("C30_KDS", 30000, 1, 2400)
                
                # Steam cured cement
                CompStrength.KDS("C25_KDS_Steam", 25000, 2, 2350)
                ```
            """
            if id == None: id =0
            js = {
                "NAME": name,
                "TYPE": "CODE",
                "CODENAME": "KDS-2016",
                "STRENGTH": fck,
                "iCTYPE": cement_type,
                "DENSITY": density
            }
            temp = CompStrength(js, id)
            self.ID = temp.ID
            self.DATA = js

    # ---------------------------------  KCI-USD12 Compressive Strength --------------------------------------------------------------

    class KCI:
        def __init__(self, name: str, fck: float = 0, cement_type: int = 1, 
                     id: int = None):
            """
            KDS-2016 Compressive Strength model.
            
            Parameters:
                name: Name
                fck: Compression strength (number) - 28-day compressive strength 
                cement_type: Cement type (integer) - Type of cement
                    • 1 - N,R moist cured: 0.35 (default)
                    • 2 - N,R steam cured: 0.15
                    • 3 - RS moist cured: 0.25
                    • 4 - RS steam cured: 0.12
                    • 5 - SL: 0.40
                id: Unique identifier (integer) - Auto-generated if not specified
            
            Examples:
                ```python
                CompStrength.KCI("C30_KCI", 30000, 1)
                
                ```
            """
            if id == None: id =0
            js = {
                "NAME": name,
                "TYPE": "CODE",
                "CODENAME": "KCI-USD12",
                "STRENGTH": fck,
                "iCTYPE": cement_type,
                
            }
            temp = CompStrength(js, id)
            self.ID = temp.ID
            self.DATA = js
    # ---------------------------------  Korean Standard Compressive Strength --------------------------------------------------------------

    class KoreanStandard:
        def __init__(self, name: str, fck: float = 0, factor_a: float = 1, 
                     factor_b: float = 2, id: int = None):
            """
            Korean Standard Compressive Strength model.
            
            Parameters:
                name: Material name (string) - Name identifier for the material
                fck: Compression strength (number) - 28-day compressive strength
                factor_a: Factor a (number) - Model parameter A (default 1)
                factor_b: Factor b (number) - Model parameter B (default 2)
                id: Unique identifier (integer) - Auto-generated if not specified
            
            Examples:
                ```python
                # Standard Korean model
                CompStrength.KoreanStandard("C30_KS", 30000, 1, 2)
                
                # Custom factors
                CompStrength.KoreanStandard("C25_KS_Custom", 25000, 1.1, 1.8)
                ```
            """
            if id == None: id =0
            js = {
                "NAME": name,
                "TYPE": "CODE",
                "CODENAME": "KoreanStandard",
                "STRENGTH": fck,
                "A": factor_a,
                "B": factor_b
            }
            temp = CompStrength(js, id)
            self.ID = temp.ID
            self.DATA = js


    # ---------------------------------  User Defined Compressive Strength --------------------------------------------------------------

    class UserDefined:
        def __init__(self, name: str, scale_factor: float = 1, 
                     time_data: list = None, id: int = None):
            """
            User Defined Compressive Strength model.
            
            Parameters:
                name: Material name (string) - Name identifier for the material
                scale_factor: Scale factor (number) - Scaling factor for the data (default 1)
                time_data: Function data (list of objects) - Time-dependent strength data
                    Each object should contain:
                    • TIME - Time in days (number)
                    • COMP - Compression strength (number)
                    • TENS - Tensile strength  (number)
                    • ELAST - Elastic modulus (number)
                id: Unique identifier (integer) - Auto-generated if not specified
            
            Examples:
                ```python
                # User defined with custom time data
                time_points = [
                    {"TIME": 0, "COMP": 0, "TENS": 0, "ELAST": 0},
                    {"TIME": 7, "COMP": 20000, "TENS": 800, "ELAST": 25000000},
                    {"TIME": 28, "COMP": 30000, "TENS": 1000, "ELAST": 30000000},
                    {"TIME": 90, "COMP": 35000, "TENS": 1200, "ELAST": 32000000}
                ]
                CompStrength.UserDefined("C30_User", 1.0, time_points)
                
                # Simple default case
                CompStrength.UserDefined("C25_User_Simple", 1.2)
                ```
            """
            if id == None: id =0
            if time_data is None:
                time_data = [
                    {"TIME": 0, "COMP": 0, "TENS": 0, "ELAST": 0},
                    {"TIME": 1000, "COMP": 30000, "TENS": 1000, "ELAST": 3000000}
                ]
            
            js = {
                "NAME": name,
                "TYPE": "USER",
                "SCALE": scale_factor,
                "aDATA": time_data
            }
            temp = CompStrength(js, id)
            self.ID = temp.ID
            self.DATA = js

#------------------------------------------ TIME DEPENDENT - MATERIAL LINK  ----------------------------------------------------



class TDMatLink:
    mats = {}
    def __init__(self,matID,CnSName='',CompName=''):

        TDMatLink.mats[str(matID)]={
            "TDMT_NAME": CnSName,
            "TDME_NAME": CompName
        }
    
    @classmethod
    def json(cls):
        json = {"Assign": TDMatLink.mats}
        return json
    
    @staticmethod
    def create():
        MidasAPI("PUT","/db/TMAT",TDMatLink.json())
        
    @staticmethod
    def get():
        return MidasAPI("GET","/db/TMAT")
    
    
    @staticmethod
    def delete():
        MidasAPI("DELETE","/db/TMAT")
        TDMatLink.clear()

    @staticmethod
    def clear():
        TDMatLink.mats={}

    @staticmethod
    def sync():
        a = TDMatLink.get()
        if a != {'message': ''}:
            if list(a['TMAT'].keys()) != []:
                TDMatLink.mats = []
                TDMatLink.ids=[]
                for j in a['TMAT'].keys():
                    TDMatLink(a['TMAT'][j], int(j))

#-------------------------------------------------------------------------------------------------


import numpy as np
# from midas_civil import *
from midas_civil import MidasAPI,Node,Element,Boundary,Thickness,Section,Model
from colorama import Fore,Style
import math
from scipy.interpolate import splev, splprep
from math import hypot
from tqdm import tqdm

# NX.version_check = False
# NX.user_print = False
# MAPI_BASEURL('http://localhost:3000/civil')
# MAPI_KEY('eyJ1ciI6InN1bWl0QG1pZGFzaXQuY29tIiwicGciOiJjaXZpbCIsImNuIjoiaXpwMnpENVVSZyJ9.d3bd3dae76e28f245958130bc44312e3ed5449f4e8d67f988b29f1f20a6055aa')
# MAPI_KEY('i_LGyGvPRjKzUB0_ZKh03A')
# NX.debug_request = True
class L2P:
    first = 0
    nDivMESH =[]
    CG_data = {}
    thick_js = {}
    sorted_nodes = []

# 1. -------------     Getting selected elements, sorting the nodes, deleting the middle node    -------------------
# Takes in [[1,2] , [3,4] ,[2,3]] returns [1,2,3,4] 
# -> MAIN FUNCTION IS TO ARRANGE THE NODE AND ELEMENT LIST
def arrangeNodeList(n_list,elm_list):
    ''' Return arranged Nodes list (1D) and Element list (1D)'''

    n_list_2 = n_list[1:]
    ord_list=[n_list[0]]

    e_list2 = elm_list[1:]
    e_ord_list=[elm_list[0]]

    add_pos = 1 # 1 => Last  | -1 => first
    flip_count=0

    def arrangeNode(ordList,n_list,add_pos,flip_count,elm_list,e_ord_list):
        
        for i in range(len(n_list)):
            if add_pos == 1:
                if ordList[-1][1] == n_list[i][0]:
                    ordList.append(n_list[i])
                    e_ord_list.append(elm_list[i])
                    del elm_list[i]
                    del n_list[i]
                    # print(f'Last add  | {ordList}')
                    return
            elif add_pos == -1:
                if ordList[0][0] == n_list[i][1]:
                    ordList.insert(0,n_list[i])
                    e_ord_list.insert(0,elm_list[i])
                    del n_list[i]
                    del elm_list[i]
                    # print(f'First add  | {ordList}')
                    return
                
        add_pos = add_pos*-1
        flip_count += 1
        # print(f'Flipped | {ordList} | {n_list} | {add_pos}')
        if flip_count == 1 : arrangeNode(ordList,n_list,add_pos,flip_count,elm_list,e_ord_list)

    for _ in range(len(n_list_2)):
        arrangeNode(ord_list,n_list_2,add_pos,flip_count,e_list2,e_ord_list)

    if len(ord_list) < len(n_list):
        print('⚠️  Element not in a single continuous line | Smaller segment is returned')
    
    simple_ord_list = [ord_list[0][0],ord_list[0][1]]
    for i in range(len(ord_list)-1):
        simple_ord_list.append(ord_list[i+1][1])

    return simple_ord_list, e_ord_list


# Returns sorted nodes and alignment coordinates for the selected elements
# -> align_nodes_list, align_coordinates_list,  beta_angle , arrangeLIST_ELM , matID , sec_ID_list_arranged,k
# SORTED NODE IDS , COORDINATES , BETA ANGLE AT NODES, SORTED ELEMENT IDS, MATID (SINGLE), SECT IDS , K - CURVE DEG(DEFAULT = 3 , MIN. BASED ON ELEMENT SELECTED)
# STEPS FOLLOWED - MIDASAPI FUNCTION IS USED
# GET THE SELECTION -> GET ELEMENT DATA ONLY FOR SELECTED ELEMENTS
# GET THE NODES OF ELEMS -> SORT THE DATA TO GET A CONTINUOUS LINE
# GET THE NODE LOCATION OF THE CONTINUOUS LINE NODES
# DELETE IN-BETWEEN NODES IF N_ELEM > 1 OR DELETE THE SELECTED ELEMENT

def delSelectElements(elemList):
    ''' 
        Deletes the middle nodes
        Returns sorted nodes and alignment coordinates for the selected elements
        Returns material ID of first element
        Returns sectID of each beam element
    '''

    if elemList == None :
        node_json = MidasAPI('GET','/view/SELECT')
        align_elem_list = node_json['SELECT']['ELEM_LIST']
    else:
        align_elem_list = elemList

    if align_elem_list == []:
        raise Exception("No elements selected ☹️")
        return 0


    align_elem_list_url = ",".join(map(str, align_elem_list))
    align_elem_json = MidasAPI('GET',f'/db/ELEM/{align_elem_list_url}')

    align_nodes_list =[]
    align_elem_list=[]


    for elmID in align_elem_json['ELEM']:
        nodes = align_elem_json['ELEM'][elmID]['NODE']
        align_nodes_list.append([nodes[0],nodes[1]])
        align_elem_list.append(elmID)

    arrangeLIST, arrangeLIST_ELM = arrangeNodeList(align_nodes_list,align_elem_list)



    matID = align_elem_json['ELEM'][arrangeLIST_ELM[0]]['MATL']

    sec_ID_list_arranged = []
    for elmID in arrangeLIST_ELM:
        sec_ID_list_arranged.append(align_elem_json['ELEM'][elmID]['SECT'])

    align_nodes_list = arrangeLIST

    align_nodes_list_url = ",".join(map(str, align_nodes_list))
    node_json = MidasAPI('GET',f'/db/NODE/{align_nodes_list_url}')


    align_coordinates_list = [] 

    for nd in align_nodes_list:
        align_coordinates_list.append([ node_json['NODE'][str(nd)]['X'] , node_json['NODE'][str(nd)]['Y'] , node_json['NODE'][str(nd)]['Z'] ])

    k=3

    if len(align_nodes_list)==2:
        MidasAPI('DELETE',f'/db/ELEM/{arrangeLIST_ELM[0]}')
        k=1

    else:
        align_nodes_list_url = ",".join(map(str, align_nodes_list[1:-1]))
        MidasAPI('DELETE',f'/db/NODE/{align_nodes_list_url}')
        k=min(3,len(align_nodes_list)-1)


    beta_angle = [align_elem_json['ELEM'][str(arrangeLIST_ELM[0])]['ANGLE']*3.141/180]
    for i in range(len(align_nodes_list)-2):
        ba1 = align_elem_json['ELEM'][str(arrangeLIST_ELM[i])]['ANGLE']*3.141/180
        ba2 = align_elem_json['ELEM'][str(arrangeLIST_ELM[i+1])]['ANGLE']*3.141/180
        beta_angle.append(0.5*ba1+0.5*ba2)
    beta_angle.append(align_elem_json['ELEM'][str(arrangeLIST_ELM[-1])]['ANGLE']*3.141/180)





    return align_nodes_list, align_coordinates_list,  beta_angle , arrangeLIST_ELM , matID , sec_ID_list_arranged,k


# 2. -------------     PLATE SECTION DEFINITION - Creating a uniform section (beta angle)   -------------------

# Orient SECTION POINT as per PROVIDED PLANE (USES NUMPY | CAN BE REPLACED WITH SCIPY)
# RETURNS TRANSFORMED POINT
def _orientPoint(plane_ax,plane_og,coord):
    #Plane orient

    Y_new = np.array(plane_ax[1])
    Z_new = np.array(plane_ax[0])
    X_new = np.cross(Y_new, Z_new)

    Y_new = np.cross(Z_new, X_new) # Recomputing

    X_new = X_new / np.linalg.norm(X_new)
    Y_new = Y_new / np.linalg.norm(Y_new)
    Z_new = Z_new / np.linalg.norm(Z_new)

    # Rotation matrix: columns are new basis vectors in original system
    R = np.vstack((X_new, Y_new, Z_new))

    # Original coordinate
    n_3dCord = coord.copy()
    n_3dCord.append(0)
    v = np.array(n_3dCord)
    # Transform the vector to the new coordinate system
    p_rot = np.dot(R.T, v)

    # Origin transform
    new_cord = [ p_rot[0]+plane_og[0] , p_rot[1]+plane_og[1] , p_rot[2]+plane_og[2] ]
    return new_cord

# Create NODES FOR THE SECTION (NODE CLASS)
# RETURNS LIST OF NODE IDS
def _createSectNodes(section_cordinates,plane_axis,plane_origin,beta_ang=0):
    node_ids = []

    for cord in section_cordinates:

        X = cord[0]*math.cos(beta_ang) - cord[1]*math.sin(beta_ang)
        Y = cord[1]*math.cos(beta_ang) + cord[0]*math.sin(beta_ang)

        ord = _orientPoint(plane_axis,plane_origin,[X,Y])    # Cord is 2D ; ord is 3D
        node_ids.append(Node(ord[0],ord[1],ord[2]).ID)

    return node_ids


# CREATE PLATE ELEMENTS BETWEEN TWO LOCATION (OPTIMISATION CAN BE DONE HERE) (ELEMENT.PLATE CLASS)
# CHECK AND CREATE THICKNESS WITH OFFSET
# RETURNS LIST OF NODE IDS FOR START AND END X-SECTION
def _createTapSectPlate(section_cordinates1,section_cordinates2,sect_lineCon,thk_plate,thk_plate_off,start_plane_axis,start_plane_origin,start_beta_angle,end_plane_axis,end_plane_origin,end_beta_angle,matID):

    s_nodes = _createSectNodes(section_cordinates1,start_plane_axis,start_plane_origin,start_beta_angle)
    e_nodes = _createSectNodes(section_cordinates2,end_plane_axis,end_plane_origin,end_beta_angle)

    for i in range(len(sect_lineCon)):

        thick_id=isCreateThick(thk_plate[i],thk_plate_off[i])

        p_node = sect_lineCon[i][0]-1
        q_node = sect_lineCon[i][1]-1

        Element.Plate([s_nodes[p_node] , e_nodes[p_node] , e_nodes[q_node] , s_nodes[q_node]],3,int(matID),thick_id)

    return s_nodes, e_nodes


# CALL ABOVE FUNCTION
# HANDLES RIGID LINK CREATION
def createTapPlateAlign(align_points,t_param,beta_angle,Section4Plate,rigid_LNK = False,matID=99):
    ''' Alignment points = [ [0,0,0] , [10,1,0]  , [20,0,0], [30,1,0]],
    t_param = [0,0.1,0.2,0.3,0.5...] list used for tap section
     Local Z vector of section is assumed [0,0,1]
      Direction is assumed  '''
    
    align_num_points = len(align_points)

    align_x_vec = [np.subtract(align_points[1],align_points[0])]

    for i in range(align_num_points-2):
        align_x_vec.append(np.add(np.subtract(align_points[i+2],align_points[i+1]), np.subtract(align_points[i+1],align_points[i])))
    align_x_vec.append(np.subtract(align_points[-1],align_points[-2]))

    align_plane = [ [p, [0.0001,0,1]] for p in align_x_vec]


    for i in range(align_num_points-1):
        ti = t_param[i]
        tf = t_param[i+1]
        shp1 , thk1 , thk_off1= getTapShape(ti,Section4Plate)
        shp2, thk2, thk_off2 = getTapShape(tf,Section4Plate)

        thk_avg = np.multiply(np.add(thk1,thk2),0.5)
        thk_off_avg = np.multiply(np.add(thk_off1,thk_off2),0.5)
        snode,enode = _createTapSectPlate(shp1,shp2,Section4Plate.LINE,thk_avg,thk_off_avg,align_plane[i],align_points[i],beta_angle[i],align_plane[i+1],align_points[i+1],beta_angle[i+1],matID)

        if rigid_LNK:
            if i == 0 :
                # beamnode = Node(align_points[0][0],align_points[0][1],align_points[0][2]).ID
                # Boundary.RigidLink(beamnode,list(set(snode)),id=5)
                Boundary.RigidLink(L2P.sorted_nodes[0],list(set(snode)),id=5)
            elif i == align_num_points-2:
                # beamnode = Node(align_points[-1][0],align_points[-1][1],align_points[-1][2]).ID
                # Boundary.RigidLink(beamnode,list(set(enode)),id=5)
                Boundary.RigidLink(L2P.sorted_nodes[-1],list(set(enode)),id=5)



# 3 . ------------------  Smooth Alignment creation ------------------

def interpolateAlignment(pointsArray,betaAngle,n_seg=10,deg=3,mSize=0):
    ''' Returns point list and beta angle list'''
    pointsArray = np.array(pointsArray)
    x_p, y_p , z_p  = pointsArray[:,0] , pointsArray[:,1] , pointsArray[:,2]
    ang_p = betaAngle


    #-- Actual length ----
    dxq = np.diff(x_p)
    dyq = np.diff(y_p)
    dzq = np.diff(z_p)
    dlq=[0]
    for i in range(len(dxq)):
        dlq.append(hypot(dxq[i],dyq[i],dzq[i]))

    tck, u = splprep([x_p, y_p, z_p, ang_p], s=0, k=deg)

    u_fine = np.linspace(0, 1, 200)
    x_den, y_den, z_den, ang_den = splev(u_fine, tck)

    dx = np.diff(x_den)
    dy = np.diff(y_den)
    dz = np.diff(z_den)
    dl=[]
    for i in range(len(dx)):
        dl.append(hypot(dx[i],dy[i],dz[i]))

    cum_l = np.insert(np.cumsum(dl),0,0)
    total_l = cum_l[-1]


    if n_seg==0 or mSize!=0:
        n_seg=int(total_l/mSize)

    eq_len = np.linspace(0,total_l,n_seg+1)

    interp_u = np.interp(eq_len,cum_l,u_fine)
    interp_x, interp_y , interp_z, iterp_ang = splev(interp_u, tck)

    align_fine_points  = [ [x, y, z] for x, y, z in zip(interp_x, interp_y , interp_z) ]

    return align_fine_points,iterp_ang, interp_u , u


class plateTapSection:

    def __init__(self,SecPoint_arr, cg_arr, t_param_arr, line_conn_arr,thk_arr,thk_off_arr):
        ''' SecPoint_arr = [SP1, SP2, SP3] |  cg_arr = [[0,0],[0,0]] | t_param_arr = [0,0.5,1]  | thk_arr = [[0.25],[0.3]] '''

        for q,shape in enumerate(SecPoint_arr):
            for i in range(len(shape)):
                shape[i] = [shape[i][0]-cg_arr[q][0] , shape[i][1]-cg_arr[q][1]]


        self.POINTS = SecPoint_arr
        self.T = t_param_arr
        self.THICK = thk_arr
        self.THICK_OFF = thk_off_arr
        self.LINE = line_conn_arr
        self.TYPE = 'TAP'


def intpolatePOINTS(t,S1,t1,S2,t2):
    t_diff = t2-t1

    S1_scaled = np.multiply(S1,((t2-t)/t_diff))
    S2_scaled = np.multiply(S2,((t-t1)/t_diff))

    S_t = np.add(S1_scaled,S2_scaled)

    return S_t

def intpolateScalar(t,S1,t1,S2,t2):
    t_diff = t2-t1

    S1_scaled = S1*(t2-t)/t_diff
    S2_scaled = S2*(t-t1)/t_diff

    S_t = S1_scaled+S2_scaled

    return S_t

def getTapShape(t,plateTapSect):
    eq_chk = 0
    i_c = 0
    for i in range(len(plateTapSect.T)):
        if t == plateTapSect.T[i_c] :
            eq_chk=1
            break
        elif t > plateTapSect.T[i_c] and t < plateTapSect.T[i_c+1]:
            break
        i_c+=1
    
    thick_curr = plateTapSect.THICK[i_c] 
    thick_off_curr = plateTapSect.THICK_OFF[i_c] 

    if eq_chk:
        return plateTapSect.POINTS[i_c] , thick_curr, thick_off_curr
    else:
        ti = plateTapSect.T[i_c]
        tf = plateTapSect.T[i_c+1]

        Si = plateTapSect.POINTS[i_c]
        Sf = plateTapSect.POINTS[i_c+1]

        thi=plateTapSect.THICK[i_c]
        thf=plateTapSect.THICK[i_c+1]

        th_off_i=plateTapSect.THICK_OFF[i_c]
        th_off_f=plateTapSect.THICK_OFF[i_c+1]
    
    
        return intpolatePOINTS(t,Si,ti,Sf,tf) , intpolatePOINTS(t,thi,ti,thf,tf), intpolatePOINTS(t,th_off_i,ti,th_off_f,tf)



def isCreateThick(thick,thick_off):
    thick = round(thick,3)
    thick_off = round(thick_off,4)

    if f'{thick}+{thick_off}' in L2P.thick_js:
        return L2P.thick_js[f'{thick}+{thick_off}']
    else:
        tid= Thickness(thick,offset=thick_off,off_type='val').ID
        L2P.thick_js[f'{thick}+{thick_off}']=tid
        return tid


#----------------------------------------------------------------
def Mesh_SHAPE(shape:Section,meshSize=0.1):
    ''' Shape is a object from midas library
    Retrurns Section points (SHAPE), Thickness, CG , Line connection of plates'''

    sect_shape, sect_thk ,sect_thk_off, sect_cgs , sect_lin_con = shape._centerLine(L2P.first) # I end for first then J end
    

    sect_cg_LT , sect_cg_CC , sect_cg_RB = sect_cgs
    # print(sect_cg_LT , sect_cg_CC , sect_cg_RB)

    offset_CG1 = [L2P.CG_data[str(shape.ID)]['Y1'],-L2P.CG_data[str(shape.ID)]['Z1']]
    offset_CG2 = [-L2P.CG_data[str(shape.ID)]['Y2'],L2P.CG_data[str(shape.ID)]['Z2']]
    sect_cg_CC1 = np.add(sect_cg_LT,offset_CG1)
    sect_cg_CC2 = np.add(sect_cg_RB,offset_CG2)

    sect_cg_CC = np.multiply(0.5,np.add(sect_cg_CC1,sect_cg_CC2))

    offset_pt = shape.OFFSET.JS['OFFSET_PT']
    offset_loc = {
            'LT':  sect_cg_LT,
            'LC':  [sect_cg_LT[0],sect_cg_CC[1]],
            'LB':  [sect_cg_LT[0],sect_cg_RB[1]],
            'CT':  [sect_cg_CC[0],sect_cg_LT[1]],
            'CC':  sect_cg_CC,
            'CB':  [sect_cg_CC[0],sect_cg_RB[1]],
            'RT':  [sect_cg_RB[0],sect_cg_LT[1]],
            'RC':  [sect_cg_RB[0],sect_cg_CC[1]],
            'RB':  sect_cg_RB
        }
    sect_cg = offset_loc[offset_pt]
    # ----------------- MESH SIZER --------------------------

    n_nodes = len(sect_shape)

    if L2P.nDivMESH == []:
        for i in (range(len(sect_lin_con))):
            if isinstance(sect_thk[i],(int,float)): sect_thk[i] = [sect_thk[i],sect_thk[i]]
            if isinstance(sect_thk_off[i],(int,float)): sect_thk_off[i] = [sect_thk_off[i],sect_thk_off[i]]
            p1 = sect_shape[sect_lin_con[i][0]-1]
            p2 = sect_shape[sect_lin_con[i][1]-1]
            dis = hypot(p1[0]-p2[0],p1[1]-p2[1])
            n_div = max(int(dis/meshSize),1)
            L2P.nDivMESH.append(n_div)
            if n_div > 1 :
                i_loc = np.linspace(p1,p2,n_div+1)
                i_thk = np.linspace(sect_thk[i][0],sect_thk[i][1],n_div+1)
                i_thk_off = np.linspace(sect_thk_off[i][0],sect_thk_off[i][1],n_div+1)

                for q in range(n_div-1):
                    sect_shape.append(i_loc[q+1])

                sect_lin_con.append([n_nodes+n_div-1,int(sect_lin_con[i][1])])
                sect_thk.append([i_thk[-2],i_thk[-1]])
                sect_thk_off.append([i_thk_off[-2],i_thk_off[-1]])
                sect_lin_con[i][1] = n_nodes+1
                sect_thk[i][1] = i_thk[1]
                sect_thk_off[i][1] = i_thk_off[1]

                for q in range(n_div-2):
                    sect_lin_con.append([n_nodes+q+1,n_nodes+q+2])
                    sect_thk.append([i_thk[q+1],i_thk[q+2]])
                    sect_thk_off.append([i_thk_off[q+1],i_thk_off[q+2]])
            
            n_nodes+=n_div-1
    else:
        for i in (range(len(sect_lin_con))):
            if isinstance(sect_thk[i],(int,float)): sect_thk[i] = [sect_thk[i],sect_thk[i]]
            if isinstance(sect_thk_off[i],(int,float)): sect_thk_off[i] = [sect_thk_off[i],sect_thk_off[i]]
            p1 = sect_shape[sect_lin_con[i][0]-1]
            p2 = sect_shape[sect_lin_con[i][1]-1]
            n_div = L2P.nDivMESH[i]
            if n_div > 1 :
                i_loc = np.linspace(p1,p2,n_div+1)
                i_thk = np.linspace(sect_thk[i][0],sect_thk[i][1],n_div+1)
                i_thk_off = np.linspace(sect_thk_off[i][0],sect_thk_off[i][1],n_div+1)

                for q in range(n_div-1):
                    sect_shape.append(i_loc[q+1])

                sect_lin_con.append([n_nodes+n_div-1,int(sect_lin_con[i][1])])
                sect_thk.append([i_thk[-2],i_thk[-1]])
                sect_thk_off.append([i_thk_off[-2],i_thk_off[-1]])
                sect_lin_con[i][1] = n_nodes+1
                sect_thk[i][1] = i_thk[1]
                sect_thk_off[i][1] = i_thk_off[1]

                for q in range(n_div-2):
                    sect_lin_con.append([n_nodes+q+1,n_nodes+q+2])
                    sect_thk.append([i_thk[q+1],i_thk[q+2]])
                    sect_thk_off.append([i_thk_off[q+1],i_thk_off[q+2]])
            
            n_nodes+=n_div-1

    for i in range(len(sect_thk)):
        sect_thk[i] = (sect_thk[i][0]+sect_thk[i][1])*0.5   #Averaging the thickness and OFFSET
        sect_thk_off[i] = (sect_thk_off[i][0]+sect_thk_off[i][1])*0.5

        
    L2P.first=1
    return sect_shape, sect_thk ,sect_thk_off, sect_cg , sect_lin_con

def getCGdata():
    js = {
        "Argument": {
            "TABLE_NAME": "SUMIT_CG",
            "TABLE_TYPE": "SECTIONALL",
            "COMPONENTS" : ["ID","Cyp","Cym","Czp","Czm"]
            }
        }
    cg_points = MidasAPI('POST','/post/TABLE',js)
    L2P.CG_data = {}
    for data in cg_points['SUMIT_CG']['DATA']:
        L2P.CG_data[data[1]] = {"Y1" : float(data[3]) , "Z1" : float(data[4]) , "Y2" : float(data[2]) , "Z2" : float(data[5])}
    # print(L2P.CG_data)

def SS_create(nSeg , mSize , bRigdLnk , meshSize, elemList):
    # ORIGINAL ALIGNMENT
    pbar = tqdm(total=14,desc="Converting Line to Plate ")

    pbar.update(1)
    pbar.set_description_str("Updating Units...")
    Model.units()
    
    pbar.update(1)
    pbar.set_description_str("Deleting Elements and Nodes...")
    sorted_node_list , align_points, align_beta_angle , elm_list, matID , align_sectID_list_sorted,k = delSelectElements(elemList) # Select elements

    L2P.sorted_nodes = sorted_node_list
    # NEW SMOOTH ALIGNMENT
    fine_align_points = align_points
    fine_beta_angle = align_beta_angle

    fine_align_points, fine_beta_angle, fine_t_param, align_t_param = interpolateAlignment(align_points,align_beta_angle,nSeg,2,mSize)


    pbar.update(1)
    pbar.set_description_str("Getting Section Data...")
    Section.sync()


    pbar.update(1)
    pbar.set_description_str("Processing Sections...")

    getCGdata()

    sect_shape_arr = []
    sect_points_arr =[]
    cg_arr = []
    thk_arr =[]
    thk_off_arr =[]
    lin =[]
    
    sect_dic = {}
    sect_ids = []

    align_sectID_list_sorted.insert(0, align_sectID_list_sorted[0])  # ADD first section again to match node count
                    
    for Sid in align_sectID_list_sorted:
        for sect in Section.sect:
            if sect.ID == Sid:
                sect_shape_arr.append(sect)

    pbar.update(1)
    pbar.set_description_str("Generating Shell definition...")
    for shape in sect_shape_arr:
        sect_shape, sect_thk , sect_thk_off, sect_cg , sect_lin_con = Mesh_SHAPE(shape,meshSize)
        sect_points_arr.append(sect_shape)
        thk_arr.append(sect_thk)
        thk_off_arr.append(sect_thk_off)
        cg_arr.append(sect_cg)
        lin = sect_lin_con

    # print('- . '*20)
    # print(cg_arr)  
    # print('- . '*20)
    # print(sect_points_arr) 
    # print('- . '*20)
                
 
    pbar.update(1)
    pbar.set_description_str("Getting existing Node data...")
    Node.clear()
    # Node.sync()
    Node.ids=[Model.maxID('NODE')]
    
    pbar.update(1)
    pbar.set_description_str("Getting existing Element data...")
    Element.clear()
    Element.ids = [Model.maxID('ELEM')]

    pbar.update(1)
    pbar.set_description_str("Getting existing Thickness data...")
    Thickness.clear()
    Thickness.ids = [Model.maxID('THIK')]

    # Boundary.RigidLink
    # Boundary.RigidLink.sync()
    pbar.update(1)
    pbar.set_description_str("Creating Shell data...")
    myTapShape = plateTapSection(sect_points_arr,cg_arr,align_t_param,lin,thk_arr,thk_off_arr)
 

    createTapPlateAlign(fine_align_points,fine_t_param,fine_beta_angle,myTapShape,bRigdLnk,matID)

    pbar.update(1)
    pbar.set_description_str("Creating Nodes...")
    Node.create()
    pbar.update(1)
    pbar.set_description_str("Creating Elements...")
    Element.create()
    pbar.update(1)
    pbar.set_description_str("Creating Thickness...")
    Thickness.create()
    pbar.update(1)
    pbar.set_description_str("Creating Rigid Links...")
    Boundary.RigidLink.create()
    # print(Boundary.RigidLink.json())

    # RESET the function
    L2P.first = 0
    L2P.nDivMESH =[]
    L2P.CG_data = {}
    L2P.thick_js = {}
    L2P.sorted_nodes = []

    pbar.update(1)
    pbar.set_description_str(Fore.GREEN+"Line to Plate conversion done"+Style.RESET_ALL)

# SS_create(0,1,True,0.5)


from ._offsetSS import Offset
from ._offsetSS import _common

class _SS_COMP_PSC_I(_common):

    """ Create Standard USER DEFINED sections"""

    def __init__(self,Name='',Symm = True,Joint=[0,0,0,0,0,0,0,0,0],
                    Bc=0,tc=0,Hh=0,
                    H1=0,
                    HL1=0,HL2=0,HL21=0,HL22=0,HL3=0,HL4=0,HL41=0,HL42=0,HL5=0,
                    BL1=0,BL2=0,BL21=0,BL22=0,BL4=0,BL41=0,BL42=0,

                    HR1=0,HR2=0,HR21=0,HR22=0,HR3=0,HR4=0,HR41=0,HR42=0,HR5=0,
                    BR1=0,BR2=0,BR21=0,BR22=0,BR4=0,BR41=0,BR42=0,

                    EgdEsb =0, DgdDsb=0,Pgd=0,Psb=0,TgdTsb=0,

                    MultiModulus = False,CreepEratio=0,ShrinkEratio=0,

                    Offset:Offset=Offset.CC(),useShear=True,use7Dof=False,id:int=0):
        
        self.ID = id
        self.NAME = Name
        self.SHAPE = 'CI'
        self.TYPE = 'COMPOSITE'

        self.SYMM = bool(Symm)

        self.BC =Bc
        self.TC =tc
        self.HH =Hh

        self.MATL_ELAST = EgdEsb
        self.MATL_DENS = DgdDsb
        self.MATL_POIS_S = Pgd
        self.MATL_POIS_C = Psb
        self.MATL_THERMAL = TgdTsb
        self.USE_MULTI_ELAST = MultiModulus
        self.LONGTERM_ESEC = CreepEratio
        self.SHRINK_ESEC = ShrinkEratio


        self.J1=bool(Joint[0])
        self.JL1=bool(Joint[1])
        self.JL2=bool(Joint[2])
        self.JL3=bool(Joint[3])
        self.JL4=bool(Joint[4])

        if self.SYMM:
            self.JR1=bool(Joint[1])
            self.JR2=bool(Joint[2])
            self.JR3=bool(Joint[3])
            self.JR4=bool(Joint[4])

            self.HR1	  =	HL1
            self.HR2	  =	HL2
            self.HR21	  =	HL21
            self.HR22	  =	HL22
            self.HR3	  =	HL3
            self.HR4	  =	HL4
            self.HR41	  =	HL41
            self.HR42	  =	HL42
            self.HR5	  =	HL5

            self.BR1	  =	BL1
            self.BR2	  =	BL2
            self.BR21	  =	BL21
            self.BR22	  =	BL22
            self.BR4	  =	BL4
            self.BR41	  =	BL41
            self.BR42	  =	BL42
        else:
            self.JR1=bool(Joint[5])
            self.JR2=bool(Joint[6])
            self.JR3=bool(Joint[7])
            self.JR4=bool(Joint[8])

            self.HR1	  =	HR1
            self.HR2	  =	HR2
            self.HR21	  =	HR21
            self.HR22	  =	HR22
            self.HR3	  =	HR3
            self.HR4	  =	HR4
            self.HR41	  =	HR41
            self.HR42	  =	HR42
            self.HR5	  =	HR5

            self.BR1	  =	BR1
            self.BR2	  =	BR2
            self.BR21	  =	BR21
            self.BR22	  =	BR22
            self.BR4	  =	BR4
            self.BR41	  =	BR41
            self.BR42	  =	BR42

        self.OFFSET = Offset
        self.USESHEAR = bool(useShear)
        self.USE7DOF = bool(use7Dof)

        self.H1	  =	H1
        self.HL1	  =	HL1
        self.HL2	  =	HL2
        self.HL21	  =	HL21
        self.HL22	  =	HL22
        self.HL3	  =	HL3
        self.HL4	  =	HL4
        self.HL41	  =	HL41
        self.HL42	  =	HL42
        self.HL5	  =	HL5

        self.BL1	  =	BL1
        self.BL2	  =	BL2
        self.BL21	  =	BL21
        self.BL22	  =	BL22
        self.BL4	  =	BL4
        self.BL41	  =	BL41
        self.BL42	  =	BL42
    
    def __str__(self):
         return f'  >  ID = {self.ID}   |  PSC COMPOSITE I SECTION \nJSON = {self.toJSON()}\n'


    def toJSON(sect):
        js =  {
                "SECTTYPE": sect.TYPE,
                "SECT_NAME": sect.NAME,
                "SECT_BEFORE": {
                    "SHAPE": sect.SHAPE,
                    "SECT_I": {
                        "vSIZE_PSC_A": [sect.H1,sect.HL1,sect.HL2,sect.HL21,sect.HL22,sect.HL3,sect.HL4,sect.HL41,sect.HL42,sect.HL5],
                        "vSIZE_PSC_B": [sect.BL1,sect.BL2,sect.BL21,sect.BL22,sect.BL4,sect.BL41,sect.BL42],
                        "vSIZE_PSC_C": [sect.HR1,sect.HR2,sect.HR21,sect.HR22,sect.HR3,sect.HR4,sect.HR41,sect.HR42,sect.HR5],
                        "vSIZE_PSC_D": [sect.BR1,sect.BR2,sect.BR21,sect.BR22,sect.BR4,sect.BR41,sect.BR42]
                    },
                    "WARPING_CHK_AUTO_I": True,
                    "WARPING_CHK_AUTO_J": True,
                    "SHEAR_CHK": True,
                    "WARPING_CHK_POS_I": [[0,0,0,0,0,0],[0,0,0,0,0,0]],
                    "WARPING_CHK_POS_J": [[0,0,0,0,0,0],[0,0,0,0,0,0]],
                    "USE_AUTO_SHEAR_CHK_POS": [[True,False,True],[False,False,False]],
                    "USE_WEB_THICK_SHEAR": [[True, True,True],[False,False,False]],
                    "SHEAR_CHK_POS": [[0,0,0],[0,0,0]],
                    "USE_WEB_THICK": [True,False],
                    "WEB_THICK": [0,0],
                    "JOINT": [sect.J1,sect.JL1,sect.JL2,sect.JL3,sect.JL4,sect.JR1,sect.JR2,sect.JR3,sect.JR4],
                    "MATL_ELAST": sect.MATL_ELAST,
                    "MATL_DENS": sect.MATL_DENS,
                    "MATL_POIS_S": sect.MATL_POIS_S,
                    "MATL_POIS_C": sect.MATL_POIS_C,
                    "MATL_THERMAL": sect.MATL_THERMAL,
                    "USE_MULTI_ELAST": sect.USE_MULTI_ELAST,
                    "LONGTERM_ESEC": sect.LONGTERM_ESEC,
                    "SHRINK_ESEC": sect.SHRINK_ESEC,
                },
                "SECT_AFTER": {
                    "SLAB": [sect.BC,sect.TC,sect.HH]
                }
            }
        js['SECT_BEFORE'].update(sect.OFFSET.JS)
        js['SECT_BEFORE']['USE_SHEAR_DEFORM'] = sect.USESHEAR
        js['SECT_BEFORE']['USE_WARPING_EFFECT'] = sect.USE7DOF
        return js
    
    @staticmethod
    def _objectify(id,name,type,shape,offset,uShear,u7DOF,js):
        vA = js['SECT_BEFORE']['SECT_I']['vSIZE_PSC_A']
        vB = js['SECT_BEFORE']['SECT_I']['vSIZE_PSC_B']
        vC = js['SECT_BEFORE']['SECT_I']['vSIZE_PSC_C']
        vD = js['SECT_BEFORE']['SECT_I']['vSIZE_PSC_D']
        joint = js['SECT_BEFORE']['JOINT']
        slab = js['SECT_AFTER']['SLAB']
        secti = js['SECT_BEFORE']

        try: e1 = js['SECT_BEFORE']['LONGTERM_ESEC'] 
        except: e1 = 0
        try: e2 = js['SECT_BEFORE']['SHRINK_ESEC'] 
        except: e2 = 0


        return _SS_COMP_PSC_I(name,False,joint,
                            slab[0],slab[1],slab[2],
                            vA[0],
                            vA[1],vA[2],vA[3],vA[4],vA[5],vA[6],vA[7],vA[8],vA[9],
                            vB[0],vB[1],vB[2],vB[3],vB[4],vB[5],vB[6],
                            vC[0],vC[1],vC[2],vC[3],vC[4],vC[5],vC[6],vC[7],vC[8],
                            vD[0],vD[1],vD[2],vD[3],vD[4],vD[5],vD[6],
                            secti['MATL_ELAST'],secti['MATL_DENS'],secti['MATL_POIS_S'],secti['MATL_POIS_C'],secti['MATL_THERMAL'],
                            secti['USE_MULTI_ELAST'],e1,e2,
                            offset,uShear,u7DOF,id)
    

class _SS_COMP_STEEL_I_TYPE1(_common):

    """ Create Standard USER DEFINED sections"""

    def __init__(self,Name='',
        Bc=0,tc=0,Hh=0,
        Hw=0,B1=0,tf1=0,tw=0,B2=0,tf2=0,

        EsEc =0, DsDc=0,Ps=0,Pc=0,TsTc=0,
        MultiModulus = False,CreepEratio=0,ShrinkEratio=0,
        Offset:Offset=Offset.CC(),useShear=True,use7Dof=False,id:int=0):
                
        self.ID = id
        self.NAME = Name
        self.SHAPE = 'I'
        self.TYPE = 'COMPOSITE'

        self.BC =Bc
        self.TC =tc
        self.HH =Hh

        self.HW	 =	Hw
        self.B1	 =	B1
        self.TF1 =	tf1
        self.TW	 =	tw
        self.B2	 =	B2    
        self.TF2  =	tf2    

        self.MATL_ELAST = EsEc
        self.MATL_DENS = DsDc
        self.MATL_POIS_S = Ps
        self.MATL_POIS_C = Pc
        self.MATL_THERMAL = TsTc
        self.USE_MULTI_ELAST = MultiModulus
        self.LONGTERM_ESEC = CreepEratio
        self.SHRINK_ESEC = ShrinkEratio

        self.OFFSET = Offset
        self.USESHEAR = bool(useShear)
        self.USE7DOF = bool(use7Dof)  
    
    def __str__(self):
         return f'  >  ID = {self.ID}   |  STEEL COMPOSITE I SECTION \nJSON = {self.toJSON()}\n'


    def toJSON(sect):
        js =  {
                "SECTTYPE": sect.TYPE,
                "SECT_NAME": sect.NAME,
                "SECT_BEFORE": {
                    "SHAPE": sect.SHAPE,
                    "SECT_I": {
                        "vSIZE": [sect.HW,sect.TW,sect.B1,sect.TF1,sect.B2,sect.TF2],
                    },
 
                    "MATL_ELAST": sect.MATL_ELAST,
                    "MATL_DENS": sect.MATL_DENS,
                    "MATL_POIS_S": sect.MATL_POIS_S,
                    "MATL_POIS_C": sect.MATL_POIS_C,
                    "MATL_THERMAL": sect.MATL_THERMAL,
                    "USE_MULTI_ELAST": sect.USE_MULTI_ELAST,
                    "LONGTERM_ESEC": sect.LONGTERM_ESEC,
                    "SHRINK_ESEC": sect.SHRINK_ESEC,
                },
                "SECT_AFTER": {
                    "SLAB": [sect.BC,sect.TC,sect.HH]
                }
            }
        js['SECT_BEFORE'].update(sect.OFFSET.JS)
        js['SECT_BEFORE']['USE_SHEAR_DEFORM'] = sect.USESHEAR
        js['SECT_BEFORE']['USE_WARPING_EFFECT'] = sect.USE7DOF
        return js
    
    @staticmethod
    def _objectify(id,name,type,shape,offset,uShear,u7DOF,js):
        vS = js['SECT_BEFORE']['SECT_I']['vSIZE']
        slab = js['SECT_AFTER']['SLAB']
        secti = js['SECT_BEFORE']

        try: e1 = js['SECT_BEFORE']['LONGTERM_ESEC'] 
        except: e1 = 0
        try: e2 = js['SECT_BEFORE']['SHRINK_ESEC'] 
        except: e2 = 0


        return _SS_COMP_STEEL_I_TYPE1(name,
                            slab[0],slab[1],slab[2],
                            vS[0],vS[2],vS[3],vS[1],vS[4],vS[5],
                            secti['MATL_ELAST'],secti['MATL_DENS'],secti['MATL_POIS_S'],secti['MATL_POIS_C'],secti['MATL_THERMAL'],
                            secti['USE_MULTI_ELAST'],e1,e2,
                            offset,uShear,u7DOF,id)


from ._mapi import MidasAPI
# from ._model import *
from ._node import Node
from ._group import Group

def convList(item):
        if type(item)!=list:
            return [item]
        else:
            return item


# -----  Extend for list of nodes/elems -----

def _ADD_Support(self):
    if isinstance(self.NODE,int):
        Boundary.Support.sups.append(self)
    elif isinstance(self.NODE,list):
        for nID in self.NODE:
            Boundary.Support(nID,self.CONST,self.GROUP)


class Boundary:

    @classmethod
    def create(cls):
        """Creates Boundary elements in MIDAS Civil NX"""
        if cls.Support.sups!=[]: cls.Support.create()
        if cls.ElasticLink.links!=[]: cls.ElasticLink.create()
        if cls.RigidLink.links!=[]: cls.RigidLink.create()
        if cls.MLFC.func!=[]: cls.RigidLink.create()
        if cls.PointSpring.springs!=[]: cls.PointSpring.create()

    
    @classmethod
    def delete(cls):
        """Delets Boundary elements from MIDAS Civil NX and Python"""
        cls.Support.delete()
        cls.ElasticLink.delete()
        cls.RigidLink.delete()
        cls.MLFC.delete()
        cls.PointSpring.delete()

    @classmethod
    def clear(cls):
        """Clear Boundary elements from Python"""
        cls.Support.clear()
        cls.ElasticLink.clear()
        cls.RigidLink.clear()
        cls.MLFC.clear()
        cls.PointSpring.clear()

    @classmethod
    def sync(cls):
        """Sync Boundary elements from MIDAS Civil NX to Python"""
        cls.Support.sync()
        cls.ElasticLink.sync()
        cls.RigidLink.sync()
        cls.MLFC.sync()
        cls.PointSpring.sync()


    class Support:
        """Create Support Object in Python \n\nNode ID, Constraint, Boundary Group.  Sample: Support(3, "1110000") or Support(3, "pin").  \nValid inputs for DOF are 1s and 0s or "pin", "fix", "free" (no capital letters).  
        \nIf more than 7 characters are entered, then only first 7 characters will be considered to define constraint."""
        sups = []
        def __init__(self, node, constraint, group = ""):
            if not isinstance(constraint, str): constraint = str(constraint)
            if constraint == "pin": constraint = "111"
            if constraint == "fix": constraint = "1111111"
            if constraint == "roller": constraint = "001"
            if len(constraint) < 7: constraint = constraint + '0' * (7-len(constraint))
            if len(constraint) > 7: constraint = constraint[:7]
            string = ''.join(['1' if char != '0' else '0' for char in constraint])

            # Check if group exists, create if not
            if group != "":
                chk = 0
                a = [v['NAME'] for v in Group.Boundary.json()["Assign"].values()]
                if group in a:
                    chk = 1
                if chk == 0:
                    Group.Boundary(group)

            self.NODE = node
            self.CONST = string
            self.GROUP = group
            self.ID = len(Boundary.Support.sups) + 1
            _ADD_Support(self)
    
        @classmethod
        def json(cls):
            """Creates JSON from Supports objects defined in Python"""
            json = {"Assign":{}}
            ng = []
            for i in Boundary.Support.sups:
                if i.NODE in Node.ids:
                    json["Assign"][i.NODE] = {"ITEMS":
                            [{"ID": i.ID,
                            "CONSTRAINT":i.CONST,
                            "GROUP_NAME": i.GROUP}]
                            }
                if i.NODE not in Node.ids: ng.append(i.NODE)
            if len(ng) > 0: print("These nodes are not defined: ", ng)
            return json
        
        @staticmethod
        def create():
            """Creates Supports in MIDAS Civil NX"""
            MidasAPI("PUT","/db/cons",Boundary.Support.json())
            
        @staticmethod
        def get():
            """Get the JSON of Supports from MIDAS Civil NX"""
            return MidasAPI("GET","/db/cons")
        
        @staticmethod
        def sync():
            """Sync Supports from MIDAS Civil NX to Python"""
            a = Boundary.Support.get()
            if a != {'message': ''}:
                if list(a['CONS'].keys()) != []:
                    Boundary.Support.sups = []
                    for j in a['CONS'].keys():
                        Boundary.Support(int(j),a['CONS'][j]['ITEMS'][0]['CONSTRAINT'])
        
        @staticmethod
        def delete():
            """Delete Supports from MIDAS Civil NX and Python"""
            Boundary.Support.clear()
            return MidasAPI("DELETE","/db/cons")

        @staticmethod
        def clear():
            """Delete Supports from Python"""
            Boundary.Support.sups=[]




    #---------------------------------------------------------------------------------------------------------------
    #Class to define Elastic Links:
    class ElasticLink:

        # list to store all link instances
        links = []
        
        def __init__(self, 
                    i_node: int, 
                    j_node: int, 
                    group: str = "", 
                    link_type: str = "GEN",
                    sdx: float = 0, 
                    sdy: float = 0, 
                    sdz: float = 0, 
                    srx: float = 0, 
                    sry: float = 0, 
                    srz: float = 0, 
                    shear: bool = False, 
                    dr_y: float = 0.5, 
                    dr_z: float = 0.5, 
                    beta_angle: float = 0, 
                    dir: str = "Dy", 
                    func_id: int = 1, 
                    distance_ratio: float = 0,
                    id: int = None, ):
            """
            Elastic link. 
            Parameters:
                i_node: The first node ID
                j_node: The second node ID
                group: The group name (default "")
                link_type: Type of link (GEN, RIGID, TENS, COMP, MULTI LINEAR, SADDLE, RAIL INTERACT) (default "GEN")
                sdx: Spring stiffness in X direction (default 0)
                sdy: Spring stiffness in Y direction (default 0)
                sdz: Spring stiffness in Z direction (default 0)
                srx: Rotational stiffness around X axis (default 0)
                sry: Rotational stiffness around Y axis (default 0)
                srz: Rotational stiffness around Z axis (default 0)
                shear: Consider shear effects (default False)
                dr_y: Distance ratio for Y direction (default 0.5)
                dr_z: Distance ratio for Z direction (default 0.5)
                beta_angle: Rotation angle in degrees (default 0)
                dir: Direction for MULTI LINEAR or RAIL INTERACT links (default "Dy")
                func_id: Function ID for MULTI LINEAR or RAIL INTERACT links (default 1)
                distance_ratio: Distance ratio for MULTI LINEAR or RAIL INTERACT links (default 0)
                id: The link ID (optional)
            
            Examples:
                ```python
                # General link with all stiffness parameters
                ElasticLink(1, 2, "Group1", "GEN", 1000, 1000, 1000, 100, 100, 100)     
                # Rigid link
                ElasticLink(3, 4, "Group2",  "RIGID")
                # Tension-only link
                ElasticLink(5, 6, "Group3", "TENS", 500)
                # Compression-only link
                ElasticLink(7, 8, "Group4",  "COMP", 500)
                # Rail Track Type link
                ElasticLink(9, 10, "Group5", "RAIL INTERACT", dir="Dy", func_id=1)
                # Multi Linear Link
                ElasticLink(11, 12, "Group6", "MULTI LINEAR", dir="Dy", func_id=1)
                # Saddle type link
                ElasticLink(13, 14, "Group7", "SADDLE")
                ```
            """
            # Check if group exists, create if not
            if group != "":
                chk = 0
                a = [v['NAME'] for v in Group.Boundary.json()["Assign"].values()]
                if group in a:
                    chk = 1
                if chk == 0:
                    Group.Boundary(group)
                    
            # Validate link type
            valid_types = ["GEN", "RIGID", "TENS", "COMP", "MULTI LINEAR", "SADDLE", "RAIL INTERACT"]
            if link_type not in valid_types:
                link_type = "GEN"
                
            # Validate direction for MULTI LINEAR
            if link_type == "MULTI LINEAR":
                valid_directions = ["Dx", "Dy", "Dz", "Rx", "Ry", "Rz"]
                if dir not in valid_directions:
                    dir = "Dy"
                    
            # Validate direction for RAIL INTERACT
            if link_type == "RAIL INTERACT":
                valid_directions = ["Dy", "Dz"]
                if dir not in valid_directions:
                    dir = "Dy"
            
            self.I_NODE = i_node
            self.J_NODE = j_node
            self.GROUP_NAME = group
            self.LINK_TYPE = link_type
            self.ANGLE = beta_angle
            
            # Parameters for all link types
            self.SDx = sdx
            self.SDy = sdy
            self.SDz = sdz
            self.SRx = srx
            self.SRy = sry
            self.SRz = srz
            self.bSHEAR = shear
            self.DR_Y = dr_y
            self.DR_Z = dr_z
            
            # Parameters for MULTI LINEAR and RAIL INTERACT
            self.Direction = dir
            self.Function_ID = func_id
            self.Distance_ratio = distance_ratio
            
            # Auto-assign ID if not provided
            if id is None:
                self.ID = len(Boundary.ElasticLink.links) + 1
            else:
                self.ID = id
                
            # Add to static list
            Boundary.ElasticLink.links.append(self)
        
        @classmethod
        def json(cls):
            """
            Converts ElasticLink data to JSON format for API submission.
            Example:
                # Get the JSON data for all links
                json_data = ElasticLink.json()
                print(json_data)
            """
            data = {}
            
            for link in cls.links:
                link_data = {
                    "NODE": [link.I_NODE, link.J_NODE],
                    "LINK": link.LINK_TYPE,
                    "ANGLE": link.ANGLE,
                    "BNGR_NAME": link.GROUP_NAME
                }
                
                # Add type-specific parameters
                if link.LINK_TYPE == "GEN":
                    link_data["R_S"] = [False] * 6
                    link_data["SDR"] = [
                        link.SDx,
                        link.SDy,
                        link.SDz,
                        link.SRx,
                        link.SRy,
                        link.SRz
                    ]
                    link_data["bSHEAR"] = link.bSHEAR
                    if link.bSHEAR:
                        link_data["DR"] = [link.DR_Y, link.DR_Z]
                    else:
                        link_data["DR"] = [0.5, 0.5]
                    
                elif link.LINK_TYPE in ["TENS", "COMP"]:
                    link_data["SDR"] = [link.SDx, 0, 0, 0, 0, 0]
                    link_data["bSHEAR"] = link.bSHEAR
                    if link.bSHEAR:
                        link_data["DR"] = [link.DR_Y, link.DR_Z]
                    else:
                        link_data["DR"] = [0.5, 0.5]
                        
                elif link.LINK_TYPE == "MULTI LINEAR":
                    direction_mapping = {
                        "Dx": 0, "Dy": 1, "Dz": 2, "Rx": 3, "Ry": 4, "Rz": 5
                    }
                    link_data["DIR"] = direction_mapping.get(link.Direction, 0)
                    link_data["MLFC"] = link.Function_ID
                    link_data["DRENDI"] = link.Distance_ratio
                    
                elif link.LINK_TYPE == "RAIL INTERACT":
                    direction_mapping = {"Dy": 1, "Dz": 2}
                    link_data["DIR"] = direction_mapping.get(link.Direction, 0)
                    link_data["RLFC"] = link.Function_ID
                    link_data["bSHEAR"] = link.bSHEAR
                    if link.bSHEAR:
                        link_data["DEENDI"] = link.Distance_ratio
                    else:
                        link_data["DR"] = [0.5, 0.5]
                    
                data[link.ID] = link_data
                
            return {"Assign": data}
        
        @classmethod
        def create(cls):
            """
            Sends all ElasticLink data to Midas API.
            Example:
                ElasticLink(1, 2, "Group1", "GEN", 1000, 1000, 1000, 100, 100, 100)
                # Send to the API
                ElasticLink.create()
            """
            MidasAPI("PUT", "/db/elnk", cls.json())
        
        @classmethod
        def get(cls):
            """
            Retrieves ElasticLink data from Midas API.
            Example:
                api_data = ElasticLink.get()
                print(api_data)
            """
            return MidasAPI("GET", "/db/elnk")
        
        @classmethod
        def sync(cls):
            """
            Updates the ElasticLink class with data from the Midas API.
            Example:
                ElasticLink.sync()
            """
            cls.links = []
            a = cls.get()
            
            if a != {'message': ''}:
                for link_id, link_data in a.get("ELNK", {}).items(): 
                    sdx = sdy = sdz = srx = sry = srz = 0
                    shear = False
                    dr_y = dr_z = 0.5
                    direction = "Dy"
                    func_id = 1
                    distance_ratio = 0

                    if link_data["LINK"] == "GEN" and "SDR" in link_data:
                        sdx, sdy, sdz, srx, sry, srz = link_data["SDR"]
                        shear = link_data.get("bSHEAR")
                        if shear and "DR" in link_data:
                            dr_y, dr_z = link_data["DR"]

                    elif link_data["LINK"] in ["TENS", "COMP"] and "SDR" in link_data:
                        sdx = link_data["SDR"][0]
                        shear = link_data.get("bSHEAR")
                        if shear and "DR" in link_data:
                            dr_y, dr_z = link_data["DR"]

                    elif link_data["LINK"] == "MULTI LINEAR":
                        dir_mapping = {0: "Dx", 1: "Dy", 2: "Dz", 3: "Rx", 4: "Ry", 5: "Rz"}
                        direction = dir_mapping.get(link_data.get("DIR"), "Dy")
                        func_id = link_data.get("MLFC")
                        distance_ratio = link_data.get("DRENDI")

                    elif link_data["LINK"] == "RAIL INTERACT":
                        dir_mapping = {1: "Dy", 2: "Dz"}
                        direction = dir_mapping.get(link_data.get("DIR"), "Dy")
                        func_id = link_data.get("RLFC")
                        shear = link_data.get("bSHEAR")
                        if shear and "DEENDI" in link_data:
                            distance_ratio = link_data["DEENDI"]

                    Boundary.ElasticLink(
                        link_data["NODE"][0],
                        link_data["NODE"][1],
                        link_data.get("BNGR_NAME"),
                        int(link_id),
                        link_data["LINK"],
                        sdx, sdy, sdz, srx, sry, srz,
                        shear, dr_y, dr_z,
                        link_data.get("ANGLE"),
                        direction, func_id, distance_ratio
                    )
        
        @classmethod
        def delete(cls):
            """
            Deletes all elastic links from the database and resets the class.
            Example:sss
                ElasticLink.delete()
            """
            cls.clear()

        @classmethod
        def clear(cls):
            """
            Deletes all elastic links from the database and resets the class.
            Example:sss
                ElasticLink.delete()
            """
            cls.links = []
    #---------------------------------------------------------------------------------------------------------------


    #Class to define Rigid  Links:
    class RigidLink:

        links = []
        
        def __init__(self, 
                    master_node: int, 
                    slave_nodes: list, 
                    group: str = "", 
                    dof: int = 111111,
                    id: int = None):
            """
            Rigid link. 
            Parameters:
                master_node: The first node ID
                slave_nodes: The second node ID
                group: The group name (default "")
                dof: Fixity of link (default 111111)
                id: The link ID (optional)
            
            Examples:
                ```python
                # General link with all stiffness parameters
                RigidLink(1, [2,3], "Group1", 111000,1)
                ```
            """

            # Check if group exists, create if not
            if group != "":
                chk = 0
                a = [v['NAME'] for v in Group.Boundary.json()["Assign"].values()]
                if group in a:
                    chk = 1
                if chk == 0:
                    Group.Boundary(group)
                    
            
            self.M_NODE = master_node
            self.S_NODE = convList(slave_nodes)
            self.GROUP_NAME = group
            self.DOF = dof

            # Auto-assign ID if not provided
            if id is None:
                self.ID = len(Boundary.RigidLink.links) + 1
            else:
                self.ID = id
                
            # Add to static list
            Boundary.RigidLink.links.append(self)
        

        @classmethod
        def json(cls):
            """
            Converts RigidLink data to JSON format for API submission.
            Example:
                # Get the JSON data for all links
                json_data = RigidLink.json()
                print(json_data)
            """
            json = {"Assign": {}}
            for link in cls.links:
                if link.M_NODE not in list(json["Assign"].keys()):
                    json["Assign"][link.M_NODE] = {"ITEMS": []}

                json["Assign"][link.M_NODE]["ITEMS"].append({
                    "ID": link.ID,
                    "GROUP_NAME": link.GROUP_NAME,
                    "DOF": link.DOF,
                    "S_NODE": convList(link.S_NODE),
                })
            return json
        
        @classmethod
        def create(cls):
            """
            Sends all RigidLink data to Midas API.
                RigidLink.create()
            """
            MidasAPI("PUT", "/db/RIGD", cls.json())
        
        @classmethod
        def get(cls):
            """
            Retrieves Rigid Link data from Midas API.
            Example:
                api_data = RigidLink.get()
                print(api_data)
            """
            return MidasAPI("GET", "/db/RIGD")
        
        @classmethod
        def sync(cls):
            """
            Updates the RigidLink class with data from the Midas API.
            Example:
                RigidLink.sync()
            """
            cls.links = []
            a = cls.get()
            if a != {'message': ''}:
                for i in a['RIGD'].keys():
                    for j in range(len(a['RIGD'][i]['ITEMS'])):
                        itm = a['RIGD'][i]['ITEMS'][j]
                        Boundary.RigidLink(int(i),itm['S_NODE'],itm['GROUP_NAME'],itm['ID'],itm['DOF'])
        
        @classmethod
        def delete(cls):
            """
            Deletes all rigid links from the database and resets the class.
            Example:
                ElasticLink.delete()
            """
            cls.clear()
            return MidasAPI("DELETE", "/db/RIGD")
        
        @classmethod
        def clear(cls):
            """
            Deletes all rigid links from the database and resets the class.
            Example:
                ElasticLink.delete()
            """
            cls.links = []
    #---------------------------------------------------------------------------------------------------------------

    class MLFC:

        func = []
        _id = []

        def __init__(self, name:str, type:str='FORCE', symm:bool=True, data:list=[[0,0],[1,1]], id:int=None):
            """
            Force-Deformation Function constructor.

            Parameters:
                name (str): The name for the Force-Deformation Function.
                type (str, optional): Type of function, either "FORCE" or "MOMENT". Defaults to "FORCE".
                symm (bool, optional): Defines if the function is symmetric (True) or unsymmetric (False). Defaults to True.
                data (list[list[float]], optional): A list of [X, Y] coordinate pairs defining the function curve. Required.
                    - For "FORCE" type: X is displacement, Y is force.
                    - For "MOMENT" type: X is rotation in radians, Y is moment.
                    Defaults to [[0,0],[1,1]].
                id (int, optional): The function ID. If not provided, it will be auto-assigned.

            Examples:
                ```python
                # Create a symmetric force vs. displacement function
                Boundary.MLFC(name="MyForceFunction", type="FORCE", symm=True, data=[[0,0], [0.1, 100], [0.2, 150]])

                # Create an unsymmetric moment vs. rotation function with a specific ID
                Boundary.MLFC(name="MyMomentFunction", type="MOMENT", symm=False, data=[[0,0], [0.01, 500], [0.02, 750]], id=5)
                ```
            """
            self.NAME = name
            self.TYPE = type
            self.SYMM = symm
            self.DATA = data

            self.X = [dat[0] for dat in self.DATA]
            self.Y = [dat[1] for dat in self.DATA]

            # Auto-assign ID if not provided
            if id is None:
                if __class__._id == []:
                    self.ID = 1
                else:
                    self.ID = max(__class__._id) + 1
            else:
                self.ID = id

            __class__._id.append(self.ID)
            __class__.func.append(self)
        

        
        @classmethod
        def json(cls):
           
            json = {"Assign": {}}
            for fn in cls.func:
                json["Assign"][fn.ID]={
                    "NAME": fn.NAME,
                    "TYPE": fn.TYPE,
                    "SYMM": fn.SYMM,
                    "FUNC_ID": 0,
                    "ITEMS": []
                }
                for i in range(len(fn.X)):
                    json["Assign"][fn.ID]["ITEMS"].append({"X":fn.X[i],"Y":fn.Y[i]})
            return json
        
        @classmethod
        def create(cls):
            """
            Sends all FUNC data.
            """
            MidasAPI("PUT", "/db/MLFC", cls.json())
        
        @classmethod
        def get(cls):
            """
            Retrieves data.
            """
            return MidasAPI("GET", "/db/MLFC")
        
        @classmethod
        def sync(cls):
            
            cls.links = []
            a = cls.get()
            if a != {'message': ''}:
                for i in a['MLFC'].keys():
                    name = a['MLFC'][i]["NAME"]
                    type = a['MLFC'][i]["TYPE"]
                    symm = a['MLFC'][i]["SYMM"]
                    data = []
                    for j in (a['MLFC'][i]['ITEMS']):
                        data.append([j["X"],j["Y"]])
                    Boundary.MLFC(name,type,symm,data,int(i))
        
        @classmethod
        def delete(cls):
            """
            Deletes all func from the database and resets the class.
            """
            cls.clear()
            return MidasAPI("DELETE", "/db/MLFC")
        
        @classmethod
        def clear(cls):
            """
            Deletes all func from the database and resets the class.
            """
            cls.links = []

#--------------------------------------------------------------------------------------------------------------------------------------


    class PointSpring:
        """Create Point Spring Object in Python"""
        springs = []
        
        def __init__(self, 
                    node: int,
                    spring_type: str = "LINEAR",
                    group: str = "",
                    stiffness: list = None,
                    fixed_option: list = None,
                    damping: list = None,
                    direction: str = "Dx+",
                    normal_vector: list = None,
                    function_id: int = 1,
                    id: int = None):
            """
            Point Spring constructor.
            
            Parameters:
                node: Node ID where spring is applied
                spring_type: Type of spring ("LINEAR", "COMP", "TENS", "MULTI")
                group: Group name (default "")
                stiffness: Spring stiffness values [SDx, SDy, SDz, SRx, SRy, SRz] or single value for COMP/TENS
                fixed_option: Fixed option array [Boolean, 6] for LINEAR type
                damping: Damping values [Cx, Cy, Cz, CRx, CRy, CRz] (if provided, damping is enabled)
                direction: Direction string ("Dx+", "Dx-", "Dy+", "Dy-", "Dz+", "Dz-", "Vector")
                normal_vector: Normal vector [x, y, z] when direction is "Vector"
                function_id: Function ID for MULTI type
                id: Spring ID (optional, auto-assigned if None)
            
            Examples:
                # Linear spring
                PointSpring(1, "LINEAR", "Group1", stiffness=[1000, 1000, 1000, 100, 100, 100])
                
                # Compression only spring
                PointSpring(2, "COMP", "Group2", stiffness=5000, direction="Dz+")
                
                # Tension only spring with vector direction
                PointSpring(3, "TENS", "Group3", stiffness=3000, direction="Vector", normal_vector=[0, -1, -1])
                
                # Multi-linear spring
                PointSpring(4, "MULTI", "Group4", direction="Dz+", function_id=1)
            """
            
            # Check if group exists, create if not
            if group != "":
                chk = 0
                a = [v['NAME'] for v in Group.Boundary.json()["Assign"].values()]
                if group in a:
                    chk = 1
                if chk == 0:
                    Group.Boundary(group)
            
            # Validate spring type
            valid_types = ["LINEAR", "COMP", "TENS", "MULTI"]
            if spring_type not in valid_types:
                spring_type = "LINEAR"
            
            self.NODE = node
            self.TYPE = spring_type
            self.GROUP_NAME = group
            
            # Convert direction string to integer
            direction_map = {
                "Dx+": 0, "Dx-": 1, "Dy+": 2, "Dy-": 3, 
                "Dz+": 4, "Dz-": 5, "Vector": 6
            }
            self.DIR = direction_map.get(direction, 0)
            
            # Auto-assign ID if not provided
            if id is None:
                self.ID = len(Boundary.PointSpring.springs) + 1
            else:
                self.ID = id
            
            # Type-specific parameters
            if spring_type == "LINEAR":
                self.SDR = stiffness if stiffness else [0, 0, 0, 0, 0, 0]
                self.F_S = fixed_option if fixed_option else [False] * 6
                # Damping logic: if damping values provided, enable damping
                self.DAMPING = damping is not None and any(d != 0 for d in damping) if damping else False
                self.Cr = damping if damping else [0, 0, 0, 0, 0, 0]
                
            elif spring_type in ["COMP", "TENS"]:
                self.STIFF = stiffness if stiffness else 0
                self.DV = normal_vector if normal_vector else [0, 0, 0]
                
            elif spring_type == "MULTI":
                self.DV = normal_vector if normal_vector else [0, 0, 0]
                self.FUNCTION = function_id
            
            # Add to static list
            Boundary.PointSpring.springs.append(self)
        
        @classmethod
        def json(cls):
            """
            Converts PointSpring data to JSON format for API submission.
            """
            json_data = {"Assign": {}}
            
            for spring in cls.springs:
                spring_data = {
                    "ID": spring.ID,
                    "TYPE": spring.TYPE,
                    "GROUP_NAME": spring.GROUP_NAME
                }
                
                # Add type-specific parameters
                if spring.TYPE == "LINEAR":
                    spring_data["SDR"] = spring.SDR
                    spring_data["F_S"] = spring.F_S
                    spring_data["DAMPING"] = spring.DAMPING
                    if spring.DAMPING:
                        spring_data["Cr"] = spring.Cr
                        
                elif spring.TYPE in ["COMP", "TENS"]:
                    spring_data["STIFF"] = spring.STIFF
                    spring_data["DIR"] = spring.DIR
                    spring_data["DV"] = spring.DV
                    
                elif spring.TYPE == "MULTI":
                    spring_data["DIR"] = spring.DIR
                    spring_data["DV"] = spring.DV
                    spring_data["FUNCTION"] = spring.FUNCTION
                
                json_data["Assign"][spring.NODE] = {"ITEMS": [spring_data]}
            
            return json_data
        
        @classmethod
        def create(cls):
            """
            Sends all PointSpring data
            """
            MidasAPI("PUT", "/db/nspr", cls.json())
        
        @classmethod
        def get(cls):
            """
            Retrieves PointSpring data
            """
            return MidasAPI("GET", "/db/nspr")
        
        @classmethod
        def sync(cls):
            """
            Updates the PointSpring class with data 
            """
            cls.springs = []
            a = cls.get()
            
            if a != {'message': ''}:
                for node_id, node_data in a.get("NSPR", {}).items():
                    for item in node_data.get("ITEMS"):
                        spring_type = item.get("TYPE")
                        group_name = item.get("GROUP_NAME")
                        spring_id = item.get("ID", 1)
                        
                        # Extract type-specific parameters
                        if spring_type == "LINEAR":
                            stiffness = item.get("SDR")
                            fixed_option = item.get("F_S")
                            damping = item.get("Cr") if item.get("DAMPING", False) else None
                            
                            # Convert direction back to string
                            dir_map = {0: "Dx+", 1: "Dx-", 2: "Dy+", 3: "Dy-", 4: "Dz+", 5: "Dz-", 6: "Vector"}
                            direction_str = dir_map.get(0, "Dx+")  # Default for LINEAR
                            
                            Boundary.PointSpring(
                                int(node_id), spring_type, group_name, spring_id,
                                stiffness, fixed_option, damping, direction_str
                            )
                            
                        elif spring_type in ["COMP", "TENS"]:
                            stiffness = item.get("STIFF")
                            direction_int = item.get("DIR")
                            normal_vector = item.get("DV")
                            
                            # Convert direction back to string
                            dir_map = {0: "Dx+", 1: "Dx-", 2: "Dy+", 3: "Dy-", 4: "Dz+", 5: "Dz-", 6: "Vector"}
                            direction_str = dir_map.get(direction_int, "Dx+")
                            
                            Boundary.PointSpring(
                                int(node_id), spring_type, group_name, spring_id,
                                stiffness, None, None, direction_str, normal_vector
                            )
                            
                        elif spring_type == "MULTI":
                            direction_int = item.get("DIR")
                            normal_vector = item.get("DV")
                            function_id = item.get("FUNCTION")
                            
                            # Convert direction back to string
                            dir_map = {0: "Dx+", 1: "Dx-", 2: "Dy+", 3: "Dy-", 4: "Dz+", 5: "Dz-", 6: "Vector"}
                            direction_str = dir_map.get(direction_int, "Dx+")
                            
                            Boundary.PointSpring(
                                int(node_id), spring_type, group_name, spring_id,
                                None, None, None, direction_str, normal_vector, function_id
                            )
        
        @classmethod
        def delete(cls):
            """
            Deletes all point springs from the database and resets the class.
            """
            cls.clear()
            return MidasAPI("DELETE", "/db/nspr")
        
        @classmethod
        def clear(cls):
            """
            Deletes all point springs from the database and resets the class.
            """
            cls.springs = []


from ._mapi import MidasAPI
from ._utils import zz_add_to_dict
from math import hypot
from ._group import _add_node_2_stGroup
from typing import Literal
import numpy as np

_ptInRad_output_ = Literal['ID','Node']

def dist_tol(a,b):
    return hypot((a.X-b.X),(a.Y-b.Y),(a.Z-b.Z)) < 0.00001  #TOLERANCE BUILT IN (UNIT INDEP)

def cell(point,size=1): #SIZE OF GRID - string format
    # return str(f"{int(point.X//size)},{int(point.Y//size)},{int(point.Z//size)}")
    return str(f"{int(point.X)},{int(point.Y)},{int(point.Z)}")

# def _cellGrid():
#     [float(x.strip()) for x in list(Node.Grid.keys())split(",")]
# -------- FUNCTIONS ARE DEFINED BELOW TO RECOGNISE NODE CLASS ----------------



#5 Class to create nodes
class Node:
    """X ordinate, Y ordinate, Z ordinate, Node ID (optional). \nSample: Node(1,0,5)"""
    nodes = [] # Node object stores in a list
    ids = []    # Node IDs used for auto increment of ID and replacement of nodes
    Grid ={}    # Node object in cube grid
    __nodeDic__ = {} # Stores node object corresponding to ID (faster get with nodebyID)
    def __init__(self,x,y,z,id:int=None,group:str='',merge:bool=True):
        ''' Create Node object

            Parameters:
                x: X - ordinate of node
                y: Y - ordinate of node 
                z: Z - ordinate of node
                id: Node ID (default 0 for auto-increment)
                mat: Material property number (default 1)
                group: Structure group of the element (str or list; 'SG1' or ['SG1','SG2'])
                merge: If enabled, checks for existing nodes and return their IDs. No additional/duplicate node will be created.
            
            Examples:
                ```python
                Node(0,0,0, id =1 , group = 'Support', merge=1)
                ```
                
        '''

        if id == None: id =0
        #----------------- ORIGINAL -----------------------
    
        if Node.ids == []: 
            node_count = 1
        else:
            node_count = max(Node.ids)+1
        
        
        self.X = round(x,6)
        self.Y = round(y,6)
        self.Z = round(z,6)

        if id == 0 : self.ID = node_count
        if id != 0 : self.ID = id


        #REPLACE - No merge check
        if id in Node.ids:

            index=Node.ids.index(id)
            n_orig = Node.nodes[index]
            loc_orig = str(cell(n_orig))
            Node.Grid[loc_orig].remove(n_orig)

            loc_new = str(cell(self))
            
            zz_add_to_dict(Node.Grid,loc_new,self)
            Node.nodes[index]=self
            Node.__nodeDic__[str(id)] = self


        #CREATE NEW - Merge Check based on input
        else:
            self.AXIS = [[0,0,0],[0,0,0],[0,0,0]]
            cell_loc = str(cell(self))      

            if cell_loc in Node.Grid:

                if merge :
                    chk=0   #OPTIONAL
                    for node in Node.Grid[cell_loc]:
                        if dist_tol(self,node):
  
                            chk=1
                            self.ID=node.ID
                            self.AXIS = node.AXIS
                    if chk==0:

                        self.AXIS = [[0,0,0],[0,0,0],[0,0,0]]
                        Node.nodes.append(self)
                        Node.ids.append(self.ID)
                        Node.Grid[cell_loc].append(self)
                        

                else:

                    Node.nodes.append(self)
                    Node.ids.append(self.ID)
                    Node.Grid[cell_loc].append(self)
            else:

                Node.Grid[cell_loc]=[]
                Node.nodes.append(self)
                Node.ids.append(self.ID)
                Node.Grid[cell_loc].append(self)
            Node.__nodeDic__[str(self.ID)] = self
            
        if group !="":
            _add_node_2_stGroup(self.ID,group)

    @property
    def LOC(self):
        ''' Return X,Y,Z as a tuple'''
        return (self.X,self.Y,self.Z)

    def __str__(self):
        return f"NODE ID : {self.ID} | X:{self.X} , Y:{self.Y} , Z:{self.Z} \n{self.__dict__}"

    @classmethod
    def json(cls):
        json = {"Assign":{}}
        for i in cls.nodes:
            json["Assign"][i.ID]={"X":i.X,"Y":i.Y,"Z":i.Z}
        return json
    
    @staticmethod
    def create():
        MidasAPI("PUT","/db/NODE",Node.json())
        
    @staticmethod
    def get():
        return MidasAPI("GET","/db/NODE")
    
    @staticmethod
    def sync():
        Node.nodes=[]
        Node.ids=[]
        Node.Grid={}
        Node.__nodeDic__ = {}
        a = Node.get()
        if a != {'message': ''}:
            if list(a['NODE'].keys()) != []:
                for j in a['NODE'].keys():
                    Node(round(a['NODE'][j]['X'],6), round(a['NODE'][j]['Y'],6), round(a['NODE'][j]['Z'],6), id=int(j), group='', merge=False)


    @staticmethod
    def delete():
        MidasAPI("DELETE","/db/NODE/")
        Node.clear()
        
    @staticmethod
    def clear():
        Node.nodes=[]
        Node.ids=[]
        Node.Grid={}
        Node.__nodeDic__ = {}








# ---- GET NODE OBJECT FROM ID ----------

# def nodeByID(nodeID:int) -> Node:
#     ''' Return Node object with the input ID '''
#     for node in Node.nodes:
#         if node.ID == nodeID:
#             return node
        
#     print(f'There is no node with ID {nodeID}')
#     return None



def nodeByID(nodeID:int) -> Node:
    ''' Return Node object with the input ID '''
    try:
        return (Node.__nodeDic__[str(nodeID)])
    except:
        print(f'There is no node with ID {nodeID}')
        return None

def closestNode(point_location:list) -> Node:
    import numpy as np
    ''' Enter location to find nearest node
        list [x,y,z] => point location => Nearest node
        node object => nearest remaining node
        int => node with ID provided => nearest remaining node  
    '''
    gridStr = list(Node.Grid.keys())
    gridInt = []
    for key in gridStr:
        gridInt.append([int(x) for x in key.split(",")])

    bNode = False
    bNodeID = 0
    if isinstance(point_location,int):
        bNode = True
        bNodeID = point_location
        nodeP = nodeByID(point_location)
        point_location = (nodeP.X,nodeP.Y,nodeP.Z)
    elif isinstance(point_location,Node):
        bNode = True
        bNodeID = point_location.ID
        point_location = (point_location.X,point_location.Y,point_location.Z)
    pGridInt = [int(point_location[0]),int(point_location[1]),int(point_location[2])]
    pGridStr = f"{int(point_location[0])},{int(point_location[1])},{int(point_location[2])}"

    min_edge_dist = round(min(point_location[0]-pGridInt[0],point_location[1]-pGridInt[1],point_location[2]-pGridInt[2]),3)
    max_edge_dist = round(max(point_location[0]-pGridInt[0],point_location[1]-pGridInt[1],point_location[2]-pGridInt[2]),3)

    if min_edge_dist > 0.5 : min_edge_dist = round(1-min_edge_dist,3)
    if max_edge_dist > 0.5 : max_edge_dist = round(1-max_edge_dist,3)

    min_edge_dist = min(min_edge_dist,max_edge_dist)

    min_dist = 10000000000  #Large value for initial value
    min_node = 0
    checked_GridInt = []

    if bNode and len(Node.Grid[pGridStr]) == 1:
        gridDist = []
        for gInt in gridInt:
            gridDist.append(abs(gInt[0]-pGridInt[0])+abs(gInt[1]-pGridInt[1])+abs(gInt[2]-pGridInt[2]))
        gridDistSort = sorted(gridDist)

        nearestGridIdx = gridDist.index(gridDistSort[1])
        nearestGridInt = gridInt[nearestGridIdx]
        nearestGridStr = gridStr[nearestGridIdx]
    else:
        if pGridInt in gridInt :
            nearestGridInt = pGridInt
            nearestGridStr = pGridStr
        else :
            gridDist = []
            for gInt in gridInt:
                gridDist.append(abs(gInt[0]-pGridInt[0])+abs(gInt[1]-pGridInt[1])+abs(gInt[2]-pGridInt[2]))

            nearestGridIdx = gridDist.index(min(gridDist))
            nearestGridInt = gridInt[nearestGridIdx]
            nearestGridStr = gridStr[nearestGridIdx]

    for nd in Node.Grid[nearestGridStr]:
        dist = hypot(nd.X-point_location[0],nd.Y-point_location[1],nd.Z-point_location[2])
        if dist < min_dist and nd.ID !=bNodeID:
            min_dist = dist
            min_node = nd
    checked_GridInt.append(nearestGridInt)
    if min_dist < min_edge_dist :
        return min_node
    
    else:
        # COMBINATION POSSIBLE FOR CELLS
        minX = int(point_location[0]-min_dist)
        maxX = int(point_location[0]+min_dist)
        minY = int(point_location[1]-min_dist)
        maxY = int(point_location[1]+min_dist)
        minZ = int(point_location[2]-min_dist)
        maxZ = int(point_location[2]+min_dist)
        possible = maxX+maxY+maxZ-minX-minY-minZ
        if possible == 0:
            return min_node

        for i in np.arange(minX,maxX+1,1):
            for j in np.arange(minY,maxY+1,1):
                for k in np.arange(minZ,maxZ+1,1):
                    cgridStr = f"{i},{j},{k}"
                    cgridInt = [i,j,k]

                    if cgridInt in checked_GridInt:
                        continue
                    else:
                        if cgridInt in gridInt:
                            for nd in Node.Grid[cgridStr]:
                                dist = hypot(nd.X-point_location[0],nd.Y-point_location[1],nd.Z-point_location[2])
                                if dist < min_dist and nd.ID !=bNodeID:
                                    min_dist = dist
                                    min_node = nd
                        checked_GridInt.append(cgridInt)
        return min_node

def _ifNodeExist_(x,y,z) -> tuple:
    cell_loc = str(f"{int(x)},{int(y)},{int(z)}")
    if cell_loc in Node.Grid:
        for node in Node.Grid[cell_loc]:
            if hypot((x-node.X),(y-node.Y),(z-node.Z)) < 0.00001 :
                return True,node.ID
    return False,0


def nodesInRadius(point_location:list , radius:float=0, output :_ptInRad_output_ = 'ID',includeSelf = False)-> list:
    gridStr = list(Node.Grid.keys())

    bNode = False
    id2Remove = 0
    if isinstance(point_location,int):
        bNode = True
        id2Remove = point_location
        nodeP = nodeByID(point_location)
        point_location = (nodeP.X,nodeP.Y,nodeP.Z)
        
    elif isinstance(point_location,Node):
        bNode = True
        id2Remove = point_location.ID
        point_location = (point_location.X,point_location.Y,point_location.Z)

    if not includeSelf and not bNode:
        bNode,id2Remove = _ifNodeExist_(point_location[0],point_location[1],point_location[2])

    ifRemove = bNode and not includeSelf

    checked_GridStr = []
    close_nodes:list[int] = []
    close_nodesID:list[Node] = []


    minX = int(point_location[0]-radius)
    maxX = int(point_location[0]+radius)
    minY = int(point_location[1]-radius)
    maxY = int(point_location[1]+radius)
    minZ = int(point_location[2]-radius)
    maxZ = int(point_location[2]+radius)

    for i in np.arange(minX,maxX+1,1):
        for j in np.arange(minY,maxY+1,1):
            for k in np.arange(minZ,maxZ+1,1):
                cgridStr = f"{i},{j},{k}"
                if cgridStr in checked_GridStr:
                    # print("Grid already checked")
                    continue
                else:
                    if cgridStr in gridStr:
                        for nd in Node.Grid[cgridStr]:
                            dist = hypot(nd.X-point_location[0],nd.Y-point_location[1],nd.Z-point_location[2])
                            if dist <= radius+0.0001 :
                                close_nodes.append(nd)
                                close_nodesID.append(nd.ID)
                    checked_GridStr.append(cgridStr)

    if output == 'Node':
        if ifRemove:
            close_nodes.remove(nodeByID(id2Remove))
            return close_nodes
    if ifRemove:
        close_nodesID.remove(id2Remove)
    return close_nodesID



class NodeLocalAxis:
    skew = []
    ids = [] 

    def __init__(self,nodeID,type,angle):
        '''
        nodeID(int) : ID of the node
        axis (str) : Axis of rotation, 'X' , 'Y' , 'Z' , 'XYZ' or 'Vector'
        angle (float) : Angle of rotation if axis = 'X' , 'Y' or 'Z'  ;
        angle (list : float) = [30,0,0] if type = 'XYZ'
        angle (list : vector) -> node.AXIS = [[1,0,0],[0,1,0]] if type = 'Vector'
        '''

        self.ID = nodeID

        if nodeID in NodeLocalAxis.ids:
            index = NodeLocalAxis.ids.index(nodeID)
            intial_angle = NodeLocalAxis.skew[index].ANGLE
            if intial_angle == [[0,0,0],[0,0,0],[0,0,0]]:
                intial_angle = [[1,0,0],[0,1,0],[0,0,1]]

            if type == 'Vector':
                self.TYPE = 'VEC'
                self.VEC = angle
            elif type == 'X':
                self.TYPE = 'ANGLE'
                self.ANGLE = [angle,intial_angle[1],intial_angle[2]]
            elif type == 'Y':
                self.TYPE = 'ANGLE'
                self.ANGLE = [intial_angle[0],angle,intial_angle[2]]
            elif type == 'Z':
                self.TYPE = 'ANGLE'
                self.ANGLE = [intial_angle[0],intial_angle[1],angle]
            elif type == 'XYZ':
                self.TYPE = 'ANGLE'
                self.ANGLE = angle
            NodeLocalAxis.skew[index] = self
        else:
            if type == 'Vector':
                self.TYPE = 'VEC'
                self.VEC = angle
                self.ANGLE = [0,0,0]
            elif type == 'X':
                self.TYPE = 'ANGLE'
                self.ANGLE = [angle,0,0]
            elif type == 'Y':
                self.TYPE = 'ANGLE'
                self.ANGLE = [0,angle,0]
            elif type == 'Z':
                self.TYPE = 'ANGLE'
                self.ANGLE = [0,0,angle]
            elif type == 'XYZ':
                self.TYPE = 'ANGLE'
                self.ANGLE = angle
        
            NodeLocalAxis.skew.append(self)
            NodeLocalAxis.ids.append(self.ID)

    @classmethod
    def json(cls):
        json = {"Assign":{}}
        for i in cls.skew:
            if i.TYPE == 'ANGLE':
                json["Assign"][i.ID]={
                                    "iMETHOD": 1,
                                    "ANGLE_X": i.ANGLE[0],
                                    "ANGLE_Y": i.ANGLE[1],
                                    "ANGLE_Z": i.ANGLE[2]
                                }
            elif i.TYPE == 'VEC':
                json["Assign"][i.ID]={
                                    "iMETHOD": 3,
                                    "V1X": i.VEC[0][0],
                                    "V1Y": i.VEC[0][1],
                                    "V1Z": i.VEC[0][2],
                                    "V2X": i.VEC[1][0],
                                    "V2Y": i.VEC[1][1],
                                    "V2Z": i.VEC[1][2]
                                }
        return json
    
    @staticmethod
    def create():
        MidasAPI("PUT","/db/SKEW",NodeLocalAxis.json())

    @staticmethod
    def delete():
        MidasAPI("DELETE","/db/SKEW/")
        NodeLocalAxis.clear()

    @staticmethod
    def clear():
        NodeLocalAxis.skew=[]
        NodeLocalAxis.ids=[]

    @staticmethod
    def get():
        return MidasAPI("GET","/db/SKEW")
    
    # @staticmethod
    # def sync():
    #     NodeLocalAxis.skew=[]
    #     NodeLocalAxis.ids=[]
    #     a = NodeLocalAxis.get()
    #     if a != {'message': ''}:
    #         if list(a['NODE'].keys()) != []:

    #             for j in a['NODE'].keys():

    #                 Node(round(a['NODE'][j]['X'],6), round(a['NODE'][j]['Y'],6), round(a['NODE'][j]['Z'],6), id=int(j), group='', merge=0)